# nmgrcomm.py v. 0.02
#
# Copyright (C) 2007 Tuomo Makinen (tjam@users.sourceforge.net)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import clientsocket

CRLF = "\r\n"

class nodemanager_connector:

    def __init__ (self, ca_certs, verify_cb):
        self.__connection = None
        self.__crypto = None
        self.__connected = False
        self.__ca_certs = ca_certs
        self.__verify_cb = verify_cb
        self.__chunk_iter = 0

        # Create connection object
        self.__create_connobject()

    def __create_connobject (self):
        if self.__connection != None:
            return

        # Client object
        if clientsocket.UseSSL:
            self.__connection = clientsocket.SSLClient(self.__verify_cb, self.__ca_certs)
            self.__connection.set_delimiter = CRLF
            self.__crypto = clientsocket.crypto
        else:
            self.__connection = None

    def is_connected (self):
        return self.__connected

    def open_connection (self, addr, port):
        if self.__connection == None:
            return (1, "clientsocket.SSLClient creation failed")
        if not isinstance(addr, str):
            return (1, "No address specified")
        if not isinstance(port, int):
            return (1, "Invalid port specification")

        cstat = self.__connection.open_connection (addr, port)
        if cstat[0] == -1:
            self.__connection.close_socket ()
            return (1, cstat[1])

        self.__connected = True

        return (0,)

    def authenticate (self, uname, passwd):
        if (self.__assert_connection() > 0):
            return None

        msg = "101 " + uname + ":" + passwd
        return self.__do_message (msg) 
  
    def logout (self):
        if (self.__assert_connection() > 0):
            return None

        msg = self.__do_message ("301")
        self.__close_connection ()
        return msg

    def query_sensors (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("201")

    def select_sensor (self, sname):
        if (self.__assert_connection() > 0):
            return None

        msg = "203 " + sname
        return self.__do_message (msg)

    def query_sensor_status (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("205")

    def query_rulesets (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("211")

    def restart_sensor (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("261")

    def shutdown_sensor (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("271")

    def freeze_sensor (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("281")

    def unfreeze_sensor (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("283")

    def remove_sensor (self):
         if (self.__assert_connection() > 0):
             return None

         return self.__do_message ("291")

    def download_ruleset (self, ruleset):
        if (self.__assert_connection() > 0):
            return None

        msg = "213 " + ruleset
        return self.__do_message (msg)

    def download_configuration (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("221")

    def upload_ruleset (self, ruleset, content):
        if (self.__assert_connection() > 0):
            return None

        msg = "215 " + ruleset + " " + content
        return self.__do_message (msg)

    def upload_configuration (self, content):
        if (self.__assert_connection() > 0):
            return None

        msg = "223 " + content
        return self.__do_message (msg)

    def send_logged_data (self, logdata):
        if (self.__assert_connection() > 0):
            return None

        msg = "231 " + logdata
        return self.__do_message (msg)

    def query_events (self):
        if (self.__assert_connection() > 0):
            return None

        return self.__do_message ("251")

    def modify_event_phase (self, data):
        if (self.__assert_connection() > 0):
	    return None

        msg = "253 " + data
        return self.__do_message (msg)

    def read_sensor_log (self, args=None):
        if (self.__assert_connection() > 0):
            return None

        if self.__chunk_iter == 0:
            if args != None:
                self.__connection.write_socket("241 " + args)
            else:
                self.__connection.write_socket("241")

        chunk = self.__read_message_chunk()
        self.__chunk_iter =+ 1

        return chunk

    def read_nodemgr_log (self, args=None):
        if (self.__assert_connection() > 0):
            return None
         
        if self.__chunk_iter == 0:
            if args != None:
                self.__connection.write_socket("243 " + args)
            else:
                self.__connection.write_socket("243")

        chunk = self.__read_message_chunk()
        self.__chunk_iter =+ 1

        return chunk

    def stop_logread (self, ack=True):
        if (self.__assert_connection() > 0):
            return None

        self.__chunk_iter = 0
        if ack:
            self.__connection.write_socket("245")
            self.__read_message()

    def __assert_connection (self):
        if (self.__connection == None or
            self.__connected == False):
            return 1

        return 0

    def __read_message_chunk (self):
        chunk = self.__connection.read_socket()
        if not chunk:
            return None
        elif chunk == -1:
            return None

        return chunk

    def __read_message (self):
        workbuf = ''

        while True:
            chunk = self.__connection.read_socket()
            if not chunk:
                break
            elif chunk == -1:
                return None

            workbuf = workbuf + chunk

            if workbuf[-2:] == CRLF:
                return workbuf

        return None

    def __do_message (self, msg):
        if self.__connection.write_socket (msg) == -1:
            return None

        rval = self.__read_message ()
        if rval != None:
            # Strip delimiter
            rval = rval[:-2]
        return rval

    def __close_connection(self):
        self.__connected = False
        self.__connection.close_socket()

