#!/usr/bin/env python

# nuheclient.py v. 0.06
#
# Copyright (C) 2007 Tuomo Makinen (tjam@users.sourceforge.net)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Nuhe node manager cli client version 0.06
Copyright (C) 2007 Tuomo Makinen

Usage:
  -i  				Interactive mode
  -a  		address		Node manager IP address
  -p  		port		Node manager TCP port
  -s  		name		Sensor name
  -c		filename	File which contains username, password and server:port
                                information (newline separated)
  -f            filename        File to save the '--fetchrule' or '--fetchconf' results (no stdout)
  --time        timestamp       Time stamp for log reading in format "Sep 07 16:00:00 2007"
                                When specified log is read from that time on

  Only one of the following options can be used at a time:

  --query      			Query sensors (shows list of registered sensors) 
  --status     			Sensor status
  --rulesets   			Sensor rulesets
  --sendrule   	ruleset	        Upload ruleset
  --fetchrule  	ruleset 	Download ruleset
  --sendconf   			Upload configuration file
  --fetchconf  			Download configuration file
  --slog       			Sensor alert logviewer
  --nlog       			Node manager logviewer
  --sendlog     string		Sends user specified string to sensor for alert inspection
  --events                      Request list of active events from sensor
  --runphase    event number    Runs next phase of event
  --lockevent   event number    Locks event at current phase
  --ulockevent  event number    Unlock event
  --freeze     			Freeze sensor
  --unfreeze   			Unfreeze sensor
  --restart    			Restart sensor
  --shutdown   			Shutdown sensor
  --remove                      Remove passive sensor entry from node manager

  --version    			Shows version
  --help       			Prints this message

  example:
  # nuheclient -i -a localhost:5403"""

import sys
import os
import string
import getopt
import getpass
import stat
import md5

from OpenSSL import crypto
    
import nmgrcomm

# Return opcodes from node manager
opcodes_r = [
      "102",
      "202",
      "204",
      "206",
      "212",
      "214",
      "216",
      "222",
      "224",
      "232",
      "242",
      "244",
      "252",
      "254",
      "262",
      "272",
      "282",
      "284",
      "292",
      "302"
    ]

# Globals
Interactive = False
Port = 5403
Addr = None
Sensor = None
Operation = None
Userdata = None
Confile = None
Outputf = None
Tstamp = None

# Trusted certificates for verification purposes
if os.name == "posix":
    CA_CERTS = os.environ["HOME"] + "/.nuhecerts"
else:
    CA_CERTS = "nuhecerts"

def main():
    if (Connector.authenticate (Uname, Passwd) != None):
        if Interactive:
            do_menu()
        else:
            if F_dict.has_key(Operation):
                do_operation()
                Connector.logout()
            else:
                fatalerror("Error: I don't know what to do!\n")
    else:
       sys.stderr.write("Authentication failed\n")    

def get_name():
    while True:
        try:
            uname = raw_input("Username: ")
        except KeyboardInterrupt:
            sys.stderr.write("\n")
            sys.exit(1)
        if uname == "":
            sys.stderr.write("Username can't be empty.\n")
        else:
            return uname

def get_passwd():
    while True:
        try:
            return getpass.getpass()
        except KeyboardInterrupt:
            sys.stderr.write("\n")
            sys.exit(1)

def do_operation():
    if Operation not in ["--query", "--nlog"]:
        rval = Connector.select_sensor(Sensor)
        if (check_message(rval, False) > 0):
            return

    if Operation == "--fetchrule" or \
       Operation == "--sendlog":
        rval = F_dict[Operation](Userdata)

    elif Operation == "--runphase" or \
         Operation == "--lockevent" or \
	 Operation == "--ulockevent":
        rval = Connector.query_events()
        if rval != None:
            if (rval[:3] == "252" and len(rval) > 3):
                elist = parse_events(rval[4:])
                if elist == None:
                    print "Error in event format"
                    return
                else:
                    try:
                        evnt = string.atoi(Userdata)
                    except ValueError:
                        print "Wrong value for event number"
                        return
                    if (evnt < 1):
                        print "Event number must be greater than 1"
                        return
                    if evnt > len(elist):
                        print "Wrong event number"
                        return

		    if Operation == "--runphase":
                        msg = do_modify_event_phase(evnt, elist, "R")
                    elif Operation == "--lockevent":
		        msg = do_modify_event_phase(evnt, elist, "L")
                    elif Operation == "--ulockevent":
		        msg = do_modify_event_phase(evnt, elist, "U")

                    if msg == None:
                        return
                    else:
                        rval = Connector.modify_event_phase(msg)

    elif Operation == "--sendrule":
        rls = read_input_str("File to upload: ")
        if not rls:
            return
        content = open_file(rls)
        if content == None:
            return
        else:
            rval = F_dict[Operation](Userdata, content)

    elif Operation == "--sendconf":
        rls = read_input_str("File to upload: ")
        if not rls:
            return
        content = open_file(rls)
        if content == None:
            return
        else:
            rval = F_dict[Operation](content)
        
    else:
        if (Operation == "--slog" or
            Operation == "--nlog"):
            do_logviewer(F_dict[Operation])
            return
        else:
            rval = F_dict[Operation]()

    check_message(rval)
     
def do_menu():
    """1) Show sensors\n2) Select sensor\n3) Node manager logviewer\n\nQ)uit"""

    sensor_s = 1
    sensor_n = ''

    print "Nodemanager: " + Addr + \
          "\n\n" + do_menu.__doc__
    print "?) Help"

    while True:

        cmd = read_input_str("# ")
        if cmd == None:
            Connector.logout()
            sys.exit(0)

        if cmd == "1":
            rval = Connector.query_sensors()
            check_message (rval)
        elif cmd == "2":
            rval = Connector.query_sensors()
            check_message (rval)
            if rval[:3] == "202":
                sdict = {}
                slist = rval[4:].split()
                if len(slist) == 0:
                    print "No sensors"
                    continue

                for i in xrange(len(slist)):
                    sdict[`i + 1`] = slist[i]

                snumb = read_input_str("sensor number: ")
                if not snumb:
                  continue

                if sdict.has_key(snumb):
                    rval = Connector.select_sensor(sdict[snumb])
                    if (check_message(rval) < 1):
	                sensor_s = 0
		        sensor_n = sdict[snumb]
                        do_sensor (sensor_n)
                else:
                    print "Wrong value"
        elif cmd == "3":
            do_logviewer (Connector.read_nodemgr_log, True)
	elif cmd == "s" or cmd == "S":
	    if sensor_s < 1:
	        do_sensor(sensor_n)
            else:
	        print "Unknown command"
        elif cmd == "q" or cmd == "Q":
            Connector.logout()
            sys.exit(0)
        elif cmd == "?":
	    print "Nodemanager: " + Addr + \
	          "\n\n" + do_menu.__doc__
            if sensor_s < 1:
	        print "S)ensor menu"
            print "?) Help"
        else:
            if cmd != '':
                print "Unknown command"

def do_sensor(sname):
    """1)  Sensor status\n2)  Sensor rulesets\n3)  Download ruleset\n4)  Upload ruleset\n\
5)  Download configuration file\n6)  Upload configuration file\n7)  Sensor alert logviewer\n\
8)  Node manager logviewer\n9)  Send logged data\n10) Remove sensor\n11) Show active events\n\
12) Modify event (run next phase, lock or unlock)\n\nF)reeze sensor\nU)nfreeze sensor\nR)estart sensor\n\
S)hutdown sensor\n\nQ)uit\nP)revious menu\n?) Help"""

    print "Sensor: " + sname + \
          "\n\n" + do_sensor.__doc__

    while True:
    
        cmd = read_input_str("# ")
        if cmd == None:
            Connector.logout()
            sys.exit(0)
          
        if cmd == "1":
            rval = Connector.query_sensor_status()
        elif cmd == "2":
            rval = Connector.query_rulesets()
        elif cmd == "3":
            rdict = get_sensor_rules()
            if rdict == None:
                print "Error reading sensor rules"
                continue

            rnumb = read_input_str("Ruleset number: ")
            if not rnumb:
                continue

            if rdict.has_key(rnumb):
                rval = Connector.download_ruleset(rdict[rnumb])
                if (check_message(rval) < 1):
                    if (read_input_str("Save to file (y/n) ? ") in ["y", "Y"]):
                        fname = read_input_str("Filename: ")
                        if fname:
                            save_file(fname, rval[4:])
            else:
                print "Wrong value"
            continue
        elif cmd == "4":
            rdict = get_sensor_rules()
            if rdict == None:
                print "Error reading sensor rules"
                continue
 
            print "N)ew ruleset"
            rnumb = read_input_str("Ruleset number or N: ")
            if not rnumb:
                continue

            if rnumb == "n" or rnumb == "N":
	        rlset = read_input_str ("Ruleset name: ")
	        if not rlset:
	            continue
            else:
                if rdict.has_key(rnumb):
                    rlset = rdict[rnumb]
                else:
                    print "Ruleset does not exist"
                    continue

	    fname = read_input_str ("File to upload: ")
	    if not fname:
	        continue
            content = open_file (fname)
            if content != None:
                rval = Connector.upload_ruleset(rlset, content)
            else:
                continue                                      
        elif cmd == "5":
            rval = Connector.download_configuration()
            if (check_message(rval) < 1):
	        if (read_input_str("Save to file (y/n) ? ") in ["y", "Y"]):
	            fname = read_input_str("Filename: ")
		    if fname:
                        save_file(fname, rval[4:])
            continue
        elif cmd == "6":
	    fname = read_input_str("File to upload: ")
	    if not fname:
	        continue
            content = open_file (fname)
            if content != None:
                rval = Connector.upload_configuration(content)
            else:
                continue
        elif cmd == "7":
            do_logviewer (Connector.read_sensor_log, True)
            continue 
        elif cmd == "8":
            do_logviewer (Connector.read_nodemgr_log, True)
            continue
        elif cmd == "9":
            ldata = read_input_str(">> ")
            if not ldata:
                continue
            rval = Connector.send_logged_data(ldata)
        elif cmd == "10":
            rval = Connector.remove_sensor()
	elif cmd == "11":
	    rval = Connector.query_events()
	elif cmd == "12":
	    rval = Connector.query_events()
            if rval != None:
	        if (rval[:3] == "252" and len(rval) > 3):
	            elist = parse_events(rval[4:])
                    if elist == None:
                        print "Error in event format"
                        continue
                    else:
	                show_events(elist)

		    st = read_input_str("Event number: ")
                    if not st:
                        continue
                    else:
		        try:
                            evnt = string.atoi(st)
                        except ValueError:
                            print "Wrong value for event number"
                            continue
                        if (evnt < 1):
                            print "Event number must be greater than 1"
                            continue
			if evnt > len(elist):
			    print "Wrong event number"
			    continue
			 
                        op = None
		        while (True):
		            print "\nR)un next phase\nL)ock\nU)nlock"
			    op = read_input_str(">> ")
			    if not op:
			        break
			    else:
                                op = op.upper()
                                if op in ["R", "L", "U"]:
                                    break
                        if not op:
			    continue

                        msg = do_modify_event_phase (evnt, elist, op)
                        if msg == None:
                            continue
                        else:
                            rval = Connector.modify_event_phase(msg)

        elif cmd == "F" or cmd == "f":
            rval = Connector.freeze_sensor()
        elif cmd == "U" or cmd == "u":
            rval = Connector.unfreeze_sensor()
        elif cmd == "R" or cmd == "r":
            if (read_input_str("Restart " + sname + " (y/n) ? ") in ["y", "Y"]):
                rval = Connector.restart_sensor()
            else:
                continue
        elif cmd == "S" or cmd == "s":
            if (read_input_str("Shutdown " + sname + " (y/n) ? ") in ["y", "Y"]):
                rval = Connector.shutdown_sensor()
            else:
                continue
        elif cmd == "Q" or cmd == "q":
            Connector.logout()
            sys.exit(1)
        elif cmd == "P" or cmd == "p":
            return
        elif cmd == "?":
	    print "Sensor: " + sname + \
	          "\n\n" + do_sensor.__doc__
            continue
        else:
            if cmd != '':
                print "Unknown command"
            continue

        check_message(rval)

def get_sensor_rules():
    rval = Connector.query_rulesets()
    check_message(rval)
    if rval[:3] != "212":
        return None
    else:
        rdict = {}
        rlist = rval[4:].split()
        if len(rlist) == 0:
            return None

        for i in xrange(len(rlist)):
            rdict[`i + 1`] = rlist[i]

    return rdict

def do_modify_event_phase(evnt, elist, eop):
    # Construct payload
    evnt = evnt - 1
    if evnt < len(elist):
        # Cleaner phase is always 1
        if elist[evnt][6] == "cleaner":
	    elist[evnt][6] = "1"

        msg = elist[evnt][7] + " " + eop + ":" + \
              elist[evnt][6] + " "

        for i in xrange(len(elist[evnt][10])):
            msg += "\"" + elist[evnt][10][i] + "\""
            if (i + 1 < len(elist[evnt][10])):
                msg += " "

        return msg
    else:
        print "Wrong event number"
        return None

def read_input_str(str):
    try:
        res = raw_input(str);
    except KeyboardInterrupt:
        sys.stdout.write("\n")
        return None
    else:
        return res

def do_logviewer(logviewer_func, interactive=False):
    global Tstamp
    checkend = False
    round = 0

    if interactive:
        print "You can specify time stamp in format \"Sep 07 16:00:00 2007\" to read log\n\
from that time on or hit return to read log from beginning"
        Tstamp = read_input_str(">> ")
        if Tstamp == None:
            return

        if isinstance (Tstamp, str):
            if Tstamp.rfind("END") != -1:
                checkend = True

    while True:
        try:
            chunk = logviewer_func(Tstamp)
        except KeyboardInterrupt:
            Connector.stop_logread()
            break
        if chunk == None:
            print "Node manager has closed connection"
            sys.exit(0)
        if round < 1:
            chunk = chunk[4:]
        if checkend and chunk[-2:] == "\r\n":
            Connector.stop_logread(False)
            break
        round += 1
        sys.stdout.write(chunk)

def check_message(message, show=True):
    
    # Node manager has closed connection
    if message == None:
        print "Node manager has closed connection"
        sys.exit(0)

    if message[:3] == "505":
        print "Request rejected"
        return 1
    if message[:3] == "507":
        print "Unknown message"
        return 1
    if message[:3] == "509":
        print "Failed request"
        return 1
    if message == "252":
        print "No active events"
        return 1
    if message[:3] not in opcodes_r:
        print "Protocol error"
        return 1
    else:
        if show:
            # Sensor events
            if message[:3] == "252":
                elist = parse_events (message[4:])
                if elist == None:
                    print "Error in event format"
                    return 1
		show_events(elist)
            # List of sensor
            elif message[:3] == "202":
                slist = message[4:].split()
                for i in xrange(len(slist)):
                    print `i + 1` + ": " + slist[i]
            # List of rules
            elif message[:3] == "212":
                rlist = message[4:].split()
                for i in xrange(len(rlist)):
                    print `i + 1` + ": " + rlist[i]
            else:
                if not Interactive:
                    # Save rule/configuration to file
                    if Outputf and message[:3] in ["214", "222"]:
                        save_file (Outputf, message[4:])
                        return 0

                print message[4:]
    return 0

def show_events(elist):
    elen = len (elist)
    blancs = 1
    if (elen >= 10 and elen < 100):
        blancs = 2
    elif (elen >= 100):
        blancs = 3

    for i in xrange(elen):
        str = `i + 1` + ":"
        str += ((blancs + 2) - len(str)) * " "
        sl = len(str)
        str += "Rule: " + elist[i][7] + (20 - len(elist[i][7])) * " " + "\"" + elist[i][8] + \
               "\"\n" + sl * " " + "Alert source:             " + "\"" + elist[i][3] + \
               "\"\n" + sl * " " + "Event creation time:      " + "\"" + elist[i][0] + \
               "\"\n" + sl * " " + "Last phase triggered at:  " + "\"" + elist[i][1] + \
               "\"\n" + sl * " " + "Next phase execution at:  " + "\"" + elist[i][2] + \
               "\"\n" + sl * " " + "Current event count:      " + elist[i][4]
 
        if elist[i][5] == "1":
	    lckphase = `string.atoi(elist[i][6]) - 1`
	    str += "\n" + sl * " " + "Event is locked at phase: " + lckphase
        else:
	    str += "\n" + sl * " " +  "Next phase for event:     " + elist[i][6]
        
	str += "\n" + sl * " " + "Matched content:          \"" + elist[i][9].replace("\\\"", "\"") + \
               "\"\n" + sl * " " + "Triggers:                 "

        # Trigger strings 
	for x in xrange(len(elist[i][10])):
	    str += "\"" + elist[i][10][x].replace("\\\"", "\"") + "\""
	    if (x + 1 < len(elist[i][10])):
	        str += ", "

        print str

def parse_events(data):
    elist = []

    splitdata = data.split("\t")
    for i in xrange(len(splitdata)):
        if splitdata[i] == "":
            del splitdata[i]

    for i in xrange(len(splitdata)):
        message = splitdata[i]

        # Creation time of event
        ix1 = message.find("itime")
        if ix1 == -1:
            return None
        ix2 = message.find("\"", ix1 + 1)
        t_dsc, s, e = return_str_between(message[ix2:], "\"", ["\\"])

        # Time when last phase was triggered
        ix1 = message.find("ptime", e + 1)
        if ix1 == -1:
            return None
        ix2 = message.find("\"", ix1 + 1)
        p_dsc, s, e = return_str_between(message[ix2:], "\"", ["\\"])

        # Time when next phase will be runned
        ix1 = message.find("etime", e + 1)
        if ix1 == -1:
            return None
        ix2 = message.find("\"", ix1 + 1)
        e_dsc, s, e = return_str_between(message[ix2:], "\"", ["\\"])

        # Alert source
        ix1 = message.find("alertsrc", e + 1)
        if ix1 == -1:
            return None
        ix2 = message.find("\"", ix1 + 1)
        a_dsc, s, e = return_str_between(message[ix2:], "\"", ["\\"])

        # Event count
        ix1 = message.find("eventcnt", e + 1)
        ix2 = message.find(":", ix1 + 1)
        ix1 = message.find(" ", ix2 + 2)
        if ix1 == -1 or ix2 == -1:
            return None
        e_id = message[ix2 + 2:ix1]

        # Event lock status
        ix1 = message.find("locked", e + 1)
        ix2 = message.find(":", ix1 + 1)
        if ix1 == -1 or ix2 == -1:
            return None
        l_id = message[ix2 + 2:ix2 + 3]

        # Next phase
        ix1 = message.find("#", e + 1)
        ix2 = message.find(":", ix1 + 1)
        if ix1 == -1 or ix2 == -1:
            return None
        p_id = message[ix1 + 1:ix2]

        # Rule number
        ix1 = message.find("#", ix2 + 1)
        ix2 = message.find(":", ix1 + 1)
        if ix1 == -1 or ix2 == -1:
	    return None
        r_id = message[ix1 + 1:ix2]

        # Rule description
        ix1 = message.find("\"", ix2 + 1)
        if ix1 == -1:
            return None
        r_dsc, s, e = return_str_between(message[ix1:], "\"", ["\\"])

        # Matched content
        ix1 = message.find("match:", e + 1)
        if ix1 == -1:
            return None
        c_str, s, e = return_str_between(message[ix1:], "\"", ["\\"])

        # Event triggers
        ix1 = message.find("triggers:", e + 1)
        ix1 = message.find("\"", ix1 + 1)
        if ix1 == -1:
	    return None

        # Collect trigger strings
        t_lst = return_str_between(message[ix1:], "\"", ["\\"], 1)

        elist.append([t_dsc, p_dsc, e_dsc, a_dsc, e_id, l_id, \
                      p_id, r_id, r_dsc, c_str, t_lst])

    return elist

def return_str_between(str, indicator, escape=[], retlist=0):
    s = e = 0
    flag = 0
    strl = []

    for i in xrange(len(str)):
        if str[i] == indicator:
            if (i > 0 and str[i - 1] in escape):
	        continue
            if flag < 1:
	        flag = 1
	        s = i
	    elif flag > 0:
	        e = i
                if retlist > 0:
                    strl.append(str[s + 1:e])
                    flag = 0
                else:
	            break

    if retlist > 0:
        return strl

    return str[s + 1:e], s, e

def save_file(fname, content):
    try:
        fp = open(fname, "w")
    except IOError, e:
        sys.stderr.write ("Failed to save file: " + e.strerror + "\n")
        return 1

    fp.write(content)
    fp.close()
    return 0

def open_file(fname):
    try:
        fp = open(fname, "r")
    except IOError, e:
        sys.stderr.write ("Failed to open file: " + e.strerror + "\n")
        return None
 
    content = fp.read()
    fp.close()
    return content
  
def verify_cb(conn, cert, errnum, depth, ok):
    if not ok:
        print "\n** Can't verify certificate ** Error " + `errnum` + " at depth " + `depth`\
              + "\n\n Issuer = %s\n Subject = %s\n MD5 Digest = %s\n" % \
              (cert.get_issuer(), cert.get_subject(),\
	       digest_cert(crypto.dump_certificate(crypto.FILETYPE_ASN1, cert)))
        if errnum == 18:    # Self signed certificate
            try:
                valid = raw_input("Do you accept certificate [y/n]? ")
            except KeyboardInterrupt:
                ok = 0
            if valid == 'y' or valid == 'Y':
                if store_certificate(crypto.dump_certificate(crypto.FILETYPE_PEM, cert)) == -1:
                    sys.stderr.write("Can't save certificate.\n")
                sys.stdout.write("\n"); ok = 1
            else:
                ok = 0
        if not ok:
            print "\nCertificate was rejected.\n"
    return ok

def read_configuration():
    try:
        fp = open(Confile).readlines()
    except IOError, e:
        fatalerror("Error: can't open " + Confile + ", " + e[1] + "\n")
    if stat.S_IMODE(os.lstat(Confile)[stat.ST_MODE]) & 077:
        fatalerror("Error: file " + Confile + " must have no more than -rwx------ (700) permissions.\n")
    else:
        c = len(fp) - 1
        while c >= 0:
            if fp[c] == "\n":
                del fp[c]
            c = c - 1
        try:
            uname = fp[0][:-1]
            passwd = fp[1][:-1]
            pindex = string.rfind(fp[2], ":")
            addr = fp[2][:pindex]
            port = fp[2][pindex + 1:-1]
        except:
            fatalerror("Error reading configuration file!\n")
    return addr, port, uname, passwd

def digest_cert(asn1cert):
    digest = md5.new(asn1cert).digest()
    return string.upper("%02x:" * len(digest) % tuple(map(ord, digest)))[:-1]

def store_certificate(cert):
    try:
        fp = open(CA_CERTS, "a")
        fp.write(cert)
    except IOError:
        return -1

    fp.close()
    return 0

def fatalerror(error_message):
    sys.stderr.write(error_message)
    sys.exit(1)

if nmgrcomm.clientsocket.UseSSL < 1:
    fatalerror("Error: can't find pyOpenSSL module\n");


# Create connector
Connector = nmgrcomm.nodemanager_connector(CA_CERTS, verify_cb)

# Functions for command line options
F_dict = {
            "--freeze": Connector.freeze_sensor,
            "--unfreeze": Connector.unfreeze_sensor,
            "--restart": Connector.restart_sensor,
            "--shutdown": Connector.shutdown_sensor,
            "--remove": Connector.remove_sensor,
            "--status": Connector.query_sensor_status,
            "--query": Connector.query_sensors,
            "--rulesets": Connector.query_rulesets,
            "--fetchrule": Connector.download_ruleset,
            "--sendrule": Connector.upload_ruleset,
            "--fetchconf": Connector.download_configuration,
            "--sendconf": Connector.upload_configuration,
            "--slog": Connector.read_sensor_log,
            "--nlog": Connector.read_nodemgr_log,
            "--sendlog": Connector.send_logged_data,
            "--events" : Connector.query_events,
            "--runphase": Connector.modify_event_phase,
	    "--lockevent": Connector.modify_event_phase,
	    "--ulockevent": Connector.modify_event_phase
         }

# --------------------------------------------------------------------
try:
    opts, args = getopt.getopt(sys.argv[1:], 'a:p:s:c:f:i',\
    ['help', 'version', 'freeze', 'unfreeze', 'restart', 'shutdown',\
     'remove', 'sendrule=', 'fetchrule=', 'sendconf', 'fetchconf',\
     'status', 'rulesets', 'time=','slog', 'nlog', 'sendlog=', 'query',\
     'events', 'runphase=', 'lockevent=', 'ulockevent='])
except:
    fatalerror("Error: try 'nuheclient.py --help' for more information.\n")

for o, a in opts:
    if o == '-i':
       Interactive = True
    elif o == '-a':
       x = string.rfind(a, ":")
       if x != -1:
           Addr = a[:x]
           Port = a[x+1:]
       else:
           Addr = a
    elif o == '-p':
        Port = a
    elif o == '-s':
       Sensor = a
    elif o == '-c':
       Confile = a
    elif o == '-f':
       Outputf = a

    elif o == '--version':
        sys.stdout.write("Nuhemanager cli client v. 0.06\n")
        sys.exit()
    elif o == '--help':
        print __doc__
        sys.exit()
    elif o == '--time':
        Tstamp = a
    else:
        if Operation != None:
            fatalerror("Error: you can use only one operation at a time\n")

        if o == '--fetchrule' or \
           o == '--sendrule' or \
           o == '--sendlog' or \
           o == '--runphase' or \
	   o == '--lockevent' or \
	   o == '--ulockevent':
            Userdata = a

        Operation = o
     
# Validate command line options
if Outputf and Operation not in ["--fetchrule", "--fetchconf"]:
    fatalerror("Error: you can only use -f option with with '--fetchrule' or '--fetchconf'\n")

if Confile != None:
    Addr, Port, Uname, Passwd = read_configuration()

if Addr == None:
    fatalerror("Error: you must specify node manager address\n")

if type(Port) == type(''):
    try:
        Port = string.atoi(Port)
    except (ValueError, TypeError):
        fatalerror("Error: invalid port specification\n")

if not Interactive:
    if (Operation != "--query" and Sensor == None):
        if Operation != "--nlog":
            fatalerror("You must provide sensor name with -s command line option\n")

# Open connection to node manager
cn = Connector.open_connection(Addr, Port)
if (cn[0] > 0):
    fatalerror("Error: can't open connection, " +
               cn[1] + "\n")

# Collect username and password
if Confile == None:
    Uname = get_name()
    Passwd = get_passwd()

# Main
main()
   
# Exit
sys.exit()
