/* event_handler.c
 *
 *  Copyright (C) 2006,2007 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "event_handler.h"

#define BUFSZ 4096

static struct node *search_node (struct list_s *elist, struct rule *rcell, char **event_id,
                                 int size);
static int do_cmd_string (char *action, char **substr, char *cstring,
			  int substr_size);
static void do_action (struct event *ecell, char *sensorname,char *whitelistfile);
static void run_external_cmd (struct event *ecell, char *cmd, char *sensorname);
static int check_ignorelist (struct action *acell, char **event_id, int size,
			     int *index);
static int register_event (struct internal *vars, struct rule *rcell, char *alertsrc,
			   char **event_id, int size);
static void update_event_status (struct node *ecell);
static int do_write_event_s (struct list_s *elist, char *eventfile);
static int do_read_event_s (struct internal *vars);
static void free_event (struct event *payload);
static struct rule *fetch_rulenr (struct internal *vars, unsigned int rulenr);
static char * prepare_str (char *srcstr);

/* Mutex for elist */
static pthread_mutex_t event_m = PTHREAD_MUTEX_INITIALIZER;

/* Populates earr d_array with active (pending events) */
int
collect_pending_events (struct internal *vars, struct d_array *earr)
{
  struct node *current;
  struct event *payload;
  int i;

  i = 0;

  pthread_mutex_lock (&event_m);
  current = vars->elist->head;

  while (current != NULL)
    {
      payload = (struct event *) current->payload;

      /* Event is pending and active */
      if (payload->pending == 1)
        {
          struct do_action *aptr;
          struct tm *tmptr;
          time_t t_ival;
          char itime[STRBUF];
          char ptime[STRBUF];
          char etime[STRBUF];
          char *newstr;
          int index;
          int subseq;
          int ei;

          subseq = 0;
          aptr = (struct do_action *) 
                   payload->acell->run_action->element[payload->action_idx]->payload;

          index = d_array_add_element (earr);
          earr->element[index]->payload = (char *) calloc (BUFSZ, sizeof (char));
          if (earr->element[index]->payload == NULL)
            {
              logwriter ("CRITICAL: failed to collect pending events, calloc()");
              pthread_mutex_unlock (&event_m);
              return -1;
            }

          newstr = prepare_str (payload->event_id[0]);
          if (newstr == NULL)
            {
              logwriter ("CRITICAL: failed to collect pending events, calloc()");
              pthread_mutex_unlock (&event_m);
              return -1;
            }

          /* Construct event message
           *************************
           */

          /* Initialization time of event */
          tmptr = (struct tm *) localtime (&payload->t_init);
          strftime (itime, STRBUF, "%b %d %H:%M:%S %Y", tmptr);

          /* Time when last event was triggered */
          tmptr = (struct tm *) localtime (&payload->t_start);
          strftime (ptime, STRBUF, "%b %d %H:%M:%S %Y", tmptr);

          /* Time stamp when next event phase will be executed */
          if (aptr->cleaner < 1)
            {
              t_ival = (time (NULL) + aptr->t_limit) - (time (NULL) - payload->t_start);
              tmptr = (struct tm *) localtime (&t_ival);
              strftime (etime, STRBUF, "%b %d %H:%M:%S %Y", tmptr);
            }

          aptr->cleaner == 0 ?
            snprintf (earr->element[index]->payload, BUFSZ,
                      "itime \"%s\" ptime \"%s\" etime \"%s\" alertsrc \"%s\" locked: %d eventcnt: %d phase #%d:\
 rule #%d: \"%s\" match: \"%s\" triggers: \"",
                      itime, ptime, etime, payload->alertsrc, payload->locked, payload->event_cnt,
                      payload->action_idx, payload->rcell->rulenr, payload->rcell->desc, newstr) :
            snprintf (earr->element[index]->payload, BUFSZ,
                      "itime \"%s\" ptime \"%s\" etime \"N/A\" alertsrc \"%s\" locked: %d eventcnt: %d phase #cleaner:\
 rule #%d: \"%s\" match: \"%s\" triggers: \"",
                      itime, ptime, payload->alertsrc, payload->locked, payload->event_cnt,
                      payload->rcell->rulenr, payload->rcell->desc, newstr);

          free (newstr);

          /* Collect trigger strings */
          for (ei = 0; ei < payload->id_size; ei++)
            {
              if (payload->acell->usesubstr[ei] == 1)
                {
                  newstr = prepare_str (payload->event_id[ei]);
                  if (newstr == NULL)
                    {
                      logwriter ("CRITICAL: failed to collect pending events, malloc()");
                      pthread_mutex_unlock (&event_m);
                      return -1;
                    }

                  if (subseq > 0)
                    strncat (earr->element[index]->payload, "\" \"", BUFSZ);
                  strncat (earr->element[index]->payload, newstr, BUFSZ);

                  free (newstr);
                  subseq++;
                }
            }
          strncat (earr->element[index]->payload, "\"\t", BUFSZ);
          i++;
        }
      current = current->next;
    }
  pthread_mutex_unlock (&event_m);
 
  return 0;
}

/* Sets event's subsequent action phase to be runned or locks/unlocks event's current phase */
int
modify_event_phase (struct internal *vars, unsigned int rulenr, unsigned int phase,
                    char *triggers, enum modevent mode)
{
  struct rule *found;
  int rval;

  rval = 1;

  /* Fetch rule structure pointer */
  found = fetch_rulenr (vars, rulenr);
  if (found != NULL)
    {
      struct node *c_event;
      char *event_id[SUBSTRMAX];
      int slen;
      int flag;
      int i;
      int x;

      flag = 0;
      for (i = 0; i < SUBSTRMAX; i++)
        event_id[i] = NULL;

      /* Adjust x correctly */
      for (x = 0; x < SUBSTRMAX; x++)
        {
          if (found->aptr->usesubstr[x] == 1)
            break;
        }

      i = 0;
      slen = strlen (triggers);
      while (i < slen)
        {
          if (triggers[i] == '\"')
            {
              /* Escaped quotation mark in substring */
              if (slen > 0 && triggers[i - 1] == '\\')
                { }
              else
                {
                  /* Start of string */
                  if (flag < 1 && x < SUBSTRMAX)
                    {
                      flag = 1;
                      event_id[x] = &triggers[i + 1];
                      x++;
                    }
                  /* End of string */
                  else if (flag == 1)
                   {
                     flag = 0;
                     triggers[i] = '\0';
                   }
                }
            }
          i++;
        }

      /* Substitute trigger strings with quotation mark escaped versions */
      for (i = 0; i < SUBSTRMAX; i++)
        {
          if (event_id[i] != NULL)
            {
              char *newstr;
              char *cptr;

              newstr = (char *) calloc (strlen (event_id[i]) + 1, sizeof (char));
              if (newstr == NULL)
                {
                  logwriter ("CRITICAL: running subsequent phase for event failed, calloc()");
                  return -1;
                }

              cptr = &newstr[0];
              for (x = 0; x < strlen (event_id[i]); x++)
                {
                  if (x + 1 < strlen (event_id[i]))
                    {
                      if (event_id[i][x] == '\\' && event_id[i][x + 1] == '\"')
                        continue;
                    }
                  *cptr++ = event_id[i][x];      
                }

              event_id[i] = (char *) calloc (strlen (newstr) + 1, sizeof (char));
              if (newstr == NULL)
                {
                  logwriter ("CRITICAL: running subsequent phase for event failed, calloc()");
                  return -1;
                }

              strcpy (event_id[i], newstr);
              free (newstr);
            }
        }

      /* Check if event exists */
      c_event = search_node (vars->elist, found, event_id, SUBSTRMAX);
      if (c_event != NULL)
        {
          struct event *payload;
          struct do_action *aptr;

          pthread_mutex_lock (&event_m);
          payload = (struct event *) c_event->payload;
          aptr = (struct do_action *) payload->acell->run_action->
                   element[payload->action_idx]->payload;

          /* Make sure that event is in correct phase */
          if (payload->action_idx == phase)
            {
              /* Set event's next phase runnable */
              if (mode == RUNPHASE)
                {
                  if (payload->locked < 1)
                    {
                      payload->runphase = 1;
                      rval = 0;
                    }        
                }
              /* Lock event at it's current phase */
              else if (mode == LOCKPHASE)
                {
                  if (aptr->cleaner == 0)
                    {
                      if (payload->locked < 1)
                        {
                          payload->locked = 1;
                          rval = 0;
                        }
                    }
                }
              /* Unlock event */
              else if (mode == ULCKPHASE)
                {
                  if (aptr->cleaner == 0)
                    {
                      if (payload->locked > 0)
                        {
                          payload->locked = 0;
                          rval = 0;
                        }
                    }
                }

              /* Save pending event structs to file */
              do_write_event_s (vars->elist, vars->eventfile);
            }
          pthread_mutex_unlock (&event_m);
        }

      for (i = 0; i < SUBSTRMAX; i++)
        {
          if (event_id[i] != NULL)
            free (event_id[i]);
        }
    }
  return rval;
}

/* Searches elist for resetable events and resets them if allowed */
int
search_and_reset_event (struct internal *vars, char *entry)
{
  struct node *current;
  struct event *payload;
  char *substr[SUBSTRMAX];
  int  flag;
  int i;

  flag = 0;

  pthread_mutex_lock (&event_m);
  current = vars->elist->head;

  while (current != NULL)
    {
      payload = (struct event *) current->payload;
      if (payload->acell->reset != NULL)
        {
          int rc;
          int idx;

          rc = -1;

          /* Loop through action struct reset d_array and perform Pcre match
           * for reset strings. NOTE: all events which share same action struct
           * (reset strings) will be reseted if substrings are matched
           */
          for (i = 0; i < d_array_get_size (payload->acell->reset); i++)
            {
              match_pcre (payload->acell->reset->element[i]->payload,
                          entry, substr, &rc);

              if (rc > 0)
                {
                  char *rstsub[SUBSTRMAX];
                  char **event_id;

                  event_id = substr;

                  /* substrtable is initialized */
                  if (payload->acell->substrtable[0] != -1)
                    {
                      int i;

                      /* Adjust reset string substrings to rule substrings */
                      for (i = 0; i < SUBSTRMAX; i++)
                        {
                          if (payload->acell->substrtable[i] != -1 &&
                              payload->acell->substrtable[i] < SUBSTRMAX)
                            rstsub[i] = substr[payload->acell->substrtable[i]];
                          else
                            rstsub[i] = NULL;
                        }
                      event_id = rstsub;
                    }

		  /* Compare event and matched reset string substrings */
	          for (idx = 0; idx < rc; idx++)
                    {
                      if (payload->acell->usesubstr[idx] == 1 &&
                          event_id[idx] != NULL)
                        {
                          if (strcmp (payload->event_id[idx], event_id[idx]) != 0)
                            {
                              flag = 0;
                              break;
                            }
                          else
                            flag = 1;
                        }
                    }

                  /* Free substrings */
                  for (idx = 0; idx < rc; idx++)
                    {
                      if (substr[idx] != NULL)
                        free (substr[idx]);
                    }

                  if (flag)
                    {
                      /* Found event: check if event is in phase where it can be reseted */
                      if (((struct do_action *) payload->acell->run_action->
                          element[payload->action_idx]->payload)->allowrst == 1)
                        payload->pending = -1;
                      break;
                    }
                }
            }
         }
       current = current->next;
    }
  pthread_mutex_unlock (&event_m);

  if (flag)
    return 0;
  return 1;
}

/* Match is found; we ignore event, register it or update it's status */
int
send_event (struct internal *vars, struct rule *rcell, char *alertsrc, 
            char **event_id, int size)
{
  struct node *current;
  int index;

  /* First check that event_id has no same substrings that ignore list */
  if (check_ignorelist
      ((struct action *) rcell->aptr, event_id, size, &index))
    {
#if 0
      char logstr[LOGBUF];
      snprintf (logstr, LOGBUF,
		"[INFO] %s, rule #%d: action ignored based on criteria: %s",
		vars->sensorname, rcell->rulenr, event_id[index]);
      send_alarm ((const char *) logstr);
#endif
      return 1;
    }

  /* Check if event is registered */
  current = search_node (vars->elist, rcell, event_id, size);
  if (current == NULL)
    {
      /* Register new event */
      pthread_mutex_lock (&event_m);
      if (register_event (vars, rcell, alertsrc, event_id, size) < 0)
	{
          logwriter ("CRITICAL: registering new event failed, malloc()");
          pthread_mutex_unlock (&event_m);
	  return -1;
	}
      pthread_mutex_unlock (&event_m);
    }
  else
    {
      /* Check if action is configured to use multiple alert
       * sources and if match is originating from same
       * source, before updating event
       */
      if (((struct event *) current->payload)->acell->multisource &&
           strcmp (alertsrc, ((struct event *)  current->payload)->alertsrc) != 0)
        return 1;

      /* Event exists, update it's status */
      update_event_status (current);
    }

  return 0;
}

static int
register_event (struct internal *vars, struct rule *rcell, char *alertsrc,
                char **event_id, int size)
{
  struct node *new_node;
  struct event *payload;
  int i;

  /* Add new event */
  new_node = slist_append (vars->elist);

  if (new_node == NULL)
    return -1;

  new_node->payload = (void *) malloc (sizeof (struct event));
  if (new_node->payload == NULL)
    return -1;

  /* Initialize event structure */
  payload = (struct event *) new_node->payload;
  payload->t_init = time (NULL);
  payload->t_start = payload->t_init;
  payload->acell = rcell->aptr;
  payload->rcell = rcell;
  payload->id_size = size;
  payload->event_cnt = 1;
  payload->action_idx = 0;
  payload->runphase = 0;
  payload->pending = 0;
  payload->locked = 0;

  /* Event is indentified by matched substrings;
   * copy substrings to event structure
   */
  payload->event_id = (char **) malloc (size * sizeof (char *));
  if (payload->event_id == NULL)
    return -1;

  for (i = 0; i < size; i++)
    {
      payload->event_id[i] =
	(char *) calloc (strlen (event_id[i]) + 1, sizeof (char));
      if (payload->event_id[i] == NULL)
	return -1;
      memcpy (payload->event_id[i], event_id[i], strlen (event_id[i]));
    }

  payload->alertsrc = (char *) calloc (strlen (alertsrc) + 1, sizeof (char));
  if (payload->alertsrc == NULL)
    return -1;
  memcpy (payload->alertsrc, alertsrc, strlen (alertsrc));

  return 0;
}

/* Update counter of existing event */
static void
update_event_status (struct node *ecell)
{
  struct event *payload;

  pthread_mutex_lock (&event_m);
  payload = (struct event *) ecell->payload;
  payload->event_cnt++;
  pthread_mutex_unlock (&event_m);
}

/* Returns pointer to string with escaped quotation marks */
static char *
prepare_str (char *srcstr)
{
  char *retstr;
  char *rptr;
  int cnt;
  int i;

  cnt = 1;
  for (i = 0; i < strlen(srcstr); i++)
    {
      if (srcstr[i] == '\"')
        cnt++;
    }

  retstr = (char *) calloc (strlen (srcstr) + cnt, sizeof (char));
  if (retstr == NULL)
    return NULL;

  rptr = &retstr[0];

  for (i = 0; i < strlen(srcstr); i++)
    {
      if (srcstr[i] == '\"')
        *rptr++ = '\\' ;
      *rptr++ = srcstr[i];
    }

  return retstr;
}

/* Searches rule by number and return pointer to rule struct */
static struct rule *
fetch_rulenr (struct internal *vars, unsigned int rulenr)
{
  struct node *current;
  struct rule *payload;

  current = vars->rlist->head;

  while (current != NULL)
    {
      payload = (struct rule *) current->payload;
      if (payload->rulenr == rulenr)
        return payload;

      current = current->next;
    }

  return NULL;
}

/* mainloop for event handler */
void
monitor_events (struct internal *vars)
{
  struct node *current;
  struct node *previous;
  struct list_s *elist;
  struct event *payload;
  struct do_action *aptr;
  struct timeval timeout;
  time_t tm;
  int aindex;
  int del_event;

  elist = vars->elist;

  /* Load saved events from file to elist */
  pthread_mutex_lock (&event_m);
  if (do_read_event_s (vars) != 0)
    logwriter("Loading of saved events failed");
  pthread_mutex_unlock (&event_m);

  while (1)
    {

#ifdef NODEMGR_SUPPORT
      /* FREEZE state means that only node manager connector thread should be
       * active; wait for state to change
       *
       * Current events are preserved as they are during FREEZE state
       */
      if (vars->run_stat == FREEZE)
        {
          pthread_mutex_lock (&freeze_sync.state_mutex);
          while (vars->run_stat == FREEZE)
            pthread_cond_wait (&freeze_sync.state_cv,
                               &freeze_sync.state_mutex);
          pthread_mutex_unlock (&freeze_sync.state_mutex);
        }
#endif

      /* quit_flag is set to exit state */
      if (quit_flag != RUNNING)
        break;

      /* Sleep for a while ... */
      timeout.tv_sec = 0;
      timeout.tv_usec = EVENT_SLEEPTIME;
      select (0, NULL, NULL, NULL, &timeout);

      pthread_mutex_lock (&event_m);
      current = elist->head;
      previous = NULL;

      /* Loop through event list */
      while (current != NULL)
	{
	  del_event = 0;
	  payload = (struct event *) current->payload;
	  aindex = payload->action_idx;
	  aptr =
	    (struct do_action *) payload->acell->run_action->element[aindex]->
	    payload;

          /* Bypass locked events */
          if (payload->locked > 0)
            goto bypass;

          /* Bypass cleaner actions if runphase is not set */
          if (aptr->cleaner == 1 && payload->runphase == 0)
            goto bypass;

          tm = time (NULL);

          /* Check if event is reseted and delete it */
          if (payload->pending == -1)
            {
              /* Save pending event structs to file */
              do_write_event_s (elist, vars->eventfile);
              del_event = 1;
            }

	  /* Only delete old actions that have positive event_max */
	  else if (aptr->event_max > 0 && payload->event_cnt <
	      aptr->event_max && (tm - payload->t_start) >= aptr->t_limit)
	    {
              /* Time elapsed without maximum amount of events */
	      del_event = 1;
	    }

	  /* Run actions if they have reached event_cnt and time limit has exceeded, also
	   * actions with event_cnt 0 will be runned based on their time limit status.
	   * Action will be runned also for pending events which has positive runphase.
	   */
          else if ((payload->runphase > 0) || (aptr->event_max > 0 && (payload->event_cnt > 0
                   && payload->event_cnt >= aptr->event_max && (tm - payload->t_start) <=
                   aptr->t_limit) || (aptr->event_max == 0 && (tm - payload->t_start) >= 
                   aptr->t_limit)))
	     {

	       /* Trigger action for this event */
	       do_action (payload, vars->sensorname,vars->whitelistfile);

	       /* Check if we have subsequent actions */
	       if (payload->action_idx < d_array_get_size (payload->acell->run_action) - 1)
 		 {
		   if (((struct do_action *) payload->acell->run_action->
		        element[aindex + 1]->payload)->event_max > 0)
                     {
                       /* We're not interested in events; subsequent action will be
                        * runned after specified time
                        */
	               payload->event_cnt = 0;
                     }

		   payload->t_start = time (NULL);
		   payload->action_idx++;
                   /* Mark this event to be pending */
                   payload->pending = 1;
		 }
	       else
		 {
		   /* No subsequent actions; delete event from list */
                   payload->pending = 0;
		   del_event = 1;
		 }
               
               /* Reset runphase */
               if (payload->runphase > 0)
                 payload->runphase = 0;

               /* Save pending event structs to file */
               do_write_event_s (elist, vars->eventfile);
	     }

          /* Delete event from list */
	  if (del_event)
	    {
              /* Free event structure */
              free_event (payload);

              /* Remove node from event list */
              if (previous != NULL)
                {
                  previous->next = current->next;
                  if (previous->next == NULL)
                    {
                      elist->tail->next = previous;
                      elist->tail = previous;
                    }
                }
              else
                {
                  elist->head = current->next;
                  if (elist->head == NULL)
                    elist->tail = NULL;
                }

	      /* Free event node */
              free (current);

	      break;
	    }

bypass:
	  previous = current;
	  current = current->next;
	}
      pthread_mutex_unlock (&event_m);
    }
}

/* Frees event structure */
static void
free_event (struct event *payload)
{
  if (payload->event_id != NULL)
    {
      int i;

      /* Free event indentification strings */
      for (i = 0; i < payload->id_size; i++)
        {
          if (payload->event_id[i] != NULL)
            free (payload->event_id[i]);
        }
      /* Free event_id */
      free (payload->event_id);
    }

  /* Free alertsrc and event struct */
  if (payload->alertsrc != NULL)
    free (payload->alertsrc);

  free (payload);
}

/* Write pending events to file */
static int
do_write_event_s (struct list_s *elist, char *eventfile)
{
  FILE *fp;
  struct node *current;
  struct event *payload;

  current = elist->head;

  fp = fopen(eventfile, "w");
  if (fp == NULL)
    {
      char err[STRBUF];
      
      snprintf (err, STRBUF, "Can't open %s for writing, %s",
                eventfile, strerror(errno)); 
      logwriter(err);
      return 1;
    }
  
  while (current != NULL)
    {
      payload = (struct event *) current->payload;
      /* Found pending event */
      if (payload->pending == 1)
        {
          int i;

          /* Start indicator */
          fprintf (fp, "****\n");
         
          fprintf(fp, "%u\n", payload->id_size); 
          for (i = 0; i < payload->id_size; i++)
            fprintf(fp, "%s\n", payload->event_id[i]);

          fprintf(fp, "%s\n%u\n%u\n%u\n%u\n%u\n%s\n%s\n%ld\n%ld\n",
                  payload->alertsrc,  payload->event_cnt, payload->action_idx,
                  payload->pending, payload->runphase, payload->locked, 
                  payload->rcell->desc, payload->rcell->actname,
                  (long) payload->t_init, (long) payload->t_start);
        }
      current = current->next;
    }

  fclose(fp);
  return 0;
}

/* Read saved events from file */
static int
do_read_event_s (struct internal *vars)
{
  FILE *fp;
  struct node *new_node;
  struct event *payload;
  char buffer[STRBUF];
  char *actname;
  char *ruledesc;
  unsigned int rstrc;
  int i;

  ruledesc = NULL;
  actname = NULL;
  payload = NULL;
  rstrc = 0;
  i = 0;

  fp = fopen(vars->eventfile, "r");
  if (fp == NULL)
    {
      char err[STRBUF];

      snprintf (err, STRBUF, "Can't open %s for reading, %s",
                vars->eventfile, strerror(errno));
      logwriter(err);
      return 1;
    }

  while (fgets (buffer, sizeof (buffer), fp) != NULL)
    {
      if (buffer[strlen (buffer) - 1] == '\n')
        buffer[strlen (buffer) - 1] = '\0';

      /* Found start of the event struct data */
      if (!strncmp (buffer, "****", 1))
        {
          /* Not ready from earlier event */
          if (rstrc > 0)
            {
              logwriter ("Warning: unable to initialize loaded event, skipping");
              free_event (payload);
            }

          rstrc = 1;
          i = 0;

          /* Create new event struct */
          payload = (struct event *) malloc (sizeof (struct event));
          if (payload == NULL)
            return -1;
          else
            {
              payload->event_id = NULL;
              payload->alertsrc = NULL;
              continue;
            }
        }

      if (rstrc)
        {
          /* Read id_size */
          if (i == 0)
            {
              sscanf (buffer, "%u", &payload->id_size);
              payload->event_id = (char **) malloc (payload->id_size * sizeof (char *));
              if (payload->event_id == NULL)
                return -1;
              else
                {
                  int ix;

                  for (ix = 0; ix < payload->id_size; ix++)
                    payload->event_id[ix] = NULL;
                }
            }
          /* Allocate space for event triggers */
          else if (i <= payload->id_size)
            {
              payload->event_id[i - 1] =
               (char *) calloc (strlen (buffer) + 1, sizeof (char));
              if (payload->event_id[i - 1] == NULL)
                return -1;
              memcpy (payload->event_id[i - 1], buffer, strlen (buffer));
            }
          else
           {
             /* First read alertsrc */
             if (i == payload->id_size + 1)
               {
                 payload->alertsrc = (char *) calloc (strlen (buffer) + 1, sizeof (char));
                 if (payload->alertsrc == NULL)
                   return -1;
                 memcpy (payload->alertsrc, buffer, strlen (buffer));
               }
             else if (i == payload->id_size + 2)
               sscanf (buffer, "%u", &payload->event_cnt);
             else if (i == payload->id_size + 3)
               sscanf (buffer, "%u", &payload->action_idx);
             else if (i == payload->id_size + 4)
               sscanf (buffer, "%u", &payload->pending);
             else if (i == payload->id_size + 5)
               sscanf (buffer, "%u", &payload->runphase);
             else if (i == payload->id_size + 6)
               sscanf (buffer, "%u", &payload->locked);
             else if (i == payload->id_size + 7)
               {
                 ruledesc = (char *) calloc (strlen (buffer) + 1, sizeof (char));
                 if (ruledesc == NULL)
                   return -1;
                 memcpy (ruledesc, buffer, strlen (buffer));
               }
             else if (i == payload->id_size + 8)
               {
                 actname = (char *) calloc (strlen (buffer) + 1, sizeof (char));
                 if (actname == NULL)
                   return -1;
                 memcpy (actname, buffer, strlen (buffer));
               }
             else if (i == payload->id_size + 9)
               {
                 long tmp;

                 sscanf (buffer, "%ld", &tmp);
                 payload->t_init = (time_t) tmp;
               }
             else if (i == payload->id_size + 10)
               {
                 struct node *current;
                 long tmp;
                 int out;

                 out = 0;
                 sscanf (buffer, "%ld", &tmp);
                 payload->t_start = (time_t) tmp;
                 
                 /* We are now ready; find corresponding rule and action structures */
                 current = vars->rlist->head;
                 while (current != NULL)
                   {
                     if (!strcmp(((struct rule *) current->payload)->actname, actname)
                         && !strcmp(((struct rule *) current->payload)->desc, ruledesc))
                       {
                         payload->rcell = (struct rule *) current->payload;
                         payload->acell = ((struct rule * )
                                           (struct action *) current->payload)->aptr;

                         /* Add new event to elist */
                         new_node = slist_append (vars->elist);
                         if (new_node != NULL)
                           {
                             new_node->payload = (void *) payload;
                             out = 1;
                           }
                         break;
                       }
                     current = current->next;
                   }

                 free (ruledesc);
                 free (actname);

                 /* No corresponding rule found */
                 if (!out)
                   {
                     logwriter ("Warning: unable to initialize loaded event, skipping");
                     free_event (payload);
                   }

                 rstrc = 0;
               }
            }
          i++;
        }
    }

  /* Last event is incomplete */
  if (rstrc > 0)
    {
      logwriter ("Warning: unable to initialize loaded event, skipping");
      free_event (payload);
    }

  fclose (fp);
  return 0;
}

/* Wrapper for do_cmd_string and run_external functions */
static void
do_action (struct event *ecell, char *sensorname,char *whitelistfile)
{

  FILE *fpc;
  char mycmd[512] = {0};
  int exec_action = 1;
  char cstring[BUFSZ] = { 0 };
  char alarmstr[LOGBUF];
  char logstr[LOGBUF];

  /* If logthreshold is set to EVENT, we log only when event action is
   * triggered
   */

  /* Generate alarms here */
  if (ecell->acell->ltvalue == EVENT)
    {
      char *rstr;

      if (ecell->action_idx == 0)
        {
          rstr =
             generate_alarm (ecell->rcell, ecell->alertsrc,
                             sensorname, ecell->event_id[0]);
     
          if (rstr != NULL)
            {
              snprintf (alarmstr, LOGBUF,
                        "%s\nAlert repeated %d times", rstr, ecell->event_cnt);
              send_alarm ((const char *) alarmstr);
              free (rstr);
            }
        }
    }

  if (do_cmd_string
      (((struct do_action *) ecell->acell->run_action->
	element[ecell->action_idx]->payload)->action, ecell->event_id,
       cstring, ecell->id_size) == -1)
    {
      snprintf (alarmstr, LOGBUF,
		"[FAILURE] %s, error: indexing substring out of bounds, can't execute '%s' command for rule #%d",
		sensorname, ((struct do_action *) ecell->acell->run_action->
		 element[ecell->action_idx]->payload)->action,
		ecell->rcell->rulenr);

      send_alarm ((const char *) alarmstr);
    }
  else{
    /* if whitelistning is in use, file exists */
    fpc  = fopen (whitelistfile, "r");
    if (fpc != NULL){
      exec_action = 0;
      while (fgets (mycmd, sizeof (mycmd), fpc) != NULL ){
        mycmd[strlen(mycmd) - 1] = 0;
	/* skip empty rows and rows starting with # */ 
        if (mycmd[0] != '\0' && !isspace(mycmd[0]) && mycmd[0] != '#' ){
	  /* allow any command to be run  */
          if (mycmd[0] == '*'){
            exec_action = 1;
	    break;
	  }
	  /* disallow combining commands with ; */
	  if ( strstr(cstring,";" ) == NULL){
	     /* check if command is whitelisted */
             if (strncmp(mycmd, cstring, strlen(mycmd)) == 0){
               exec_action = 1;
	       break;
             }
          }
        }
      }
    }
    fclose (fpc);

    if ( exec_action == 1){
      run_external_cmd (ecell, cstring, sensorname);
    }else{
      snprintf (logstr, LOGBUF,
                    "[FAILURE] %s, error: can't execute non whitelisted command '%s', %s",
                    sensorname, cstring,
                    strerror (errno));
      send_alarm ((const char *) logstr);
    }
      
  }
}

/* This function copies action string with modifications to cstring
 * ($0-9 characters are substituded with corresponding Pcre substrings)
 */
static int
do_cmd_string (char *action, char **substr, char *cstring, int substr_size)
{
  int i = 0;
  int x = 0;
  int y;
  int sidx;
     
  while (i < strlen (action))
    {
      if (i >= BUFSZ)
	return 0;
      if (action[i] == '$' && i < strlen (action) - 1)
	{
	  sidx = atoi (&action[++i]);
	  if (sidx > substr_size - 1)
	    return -1;
	  for (y = 0; y < strlen (substr[sidx]); y++)
	    cstring[x++] = substr[sidx][y];
	}
      else
	cstring[x++] = action[i];
      i++;
    }
  return 0;
}

static void
run_external_cmd (struct event *ecell, char *cmd, char *sensorname)
{
  FILE *fp;
  int rval;
  char logstr[LOGBUF];


  /* Check what kind of action indentifier we have */
  if (ecell->acell->id == CMD)
    {
      /* Run external command specified by rule action */

      /* FIX this to not use popen() */
      fp = popen (cmd, "r");
      if (fp == NULL)
	{
	  snprintf (logstr, LOGBUF,
		    "[FAILURE] %s, error: can't execute '%s' for rule #%d, %s", 
                    sensorname, cmd,
		    ecell->rcell->rulenr, strerror (errno));

	  send_alarm ((const char *) logstr);
	}
      else
	{
	  rval = pclose (fp);

	  snprintf (logstr, LOGBUF, "[ACTION] %s, executed action '%s' command '%s' for rule #%d '%s' with return value of %d", sensorname, ecell->rcell->actname, cmd, ecell->rcell->rulenr,ecell->rcell->desc ,rval);

	  send_alarm ((const char *) logstr);
	}
    }
}

/* This function validates if event should be ignored (based on ignorelist) */
static int
check_ignorelist (struct action *acell, char **event_id, int size, int *index)
{
  int i;
  int x;

  char *ign_data;

  for (i = 0; i < d_array_get_size (acell->ignore); i++)
    {
      for (x = 0; x < size; x++)
	{
          ign_data = (char *) acell->ignore->element[i]->payload;
	  if (strcmp (ign_data, event_id[x]) == 0)
	    {
	      *index = x;
	      return 1;
	    }
	}
    }
  return 0;
}

/* Searches event list, returns pointer to event structure if found
 * or NULL otherwise
 */
static struct node *
search_node (struct list_s *elist, struct rule *rcell, char **event_id, int size)
{
  struct node *current;
  struct event *payload;
  int i;
  int flag;

  flag = 0;
  pthread_mutex_lock (&event_m);
  current = elist->head;

  while (current != NULL)
    {
      payload = (struct event *) current->payload;

      /* Check that we have correct rule */
      if (payload->rcell->rulenr == rcell->rulenr)
        {
          for (i = 0; i < size; i++)
	    {
	      if (payload->acell->usesubstr[i] == 1 &&
                event_id[i] != NULL)
	        {
                  /* Check that we have correct event */
	          if (strcmp (payload->event_id[i], event_id[i]) != 0)
	            {
		      flag = 0;
		      break;
		    }
	          else
                    flag = 1;
	        }
	    }

          if (flag == 1)
	    {
	      /* Event found */
	      pthread_mutex_unlock (&event_m);
	      return current;
	    }
        }

      current = current->next;
    }
  pthread_mutex_unlock (&event_m);

  return NULL;
}
