/* nuhed.c
 *
 * Copyright (C) 2006 - 2023 Tuomo Makinen (tjam@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "nuhed.h"

/* Maximum amount of errors tolerated during file reopening */
#define ERRSTATMAX -5

/* Logging mode */
enum logmethod
{
  LOG_NODEMGR = 0,
  LOG_FILE,
  LOG_BOTH,
  LOG_STDOUT
};

/* Running mode */
enum runmethod
{
  STANDALONE = 0,
  NODEMANAGER
};

/* Monitored file status */
enum fstatus
{
  ACTIVE = 0,
  INACTIVE,
  REOPEN,
  DISABLED
};

/* File info structure */
struct finfo
{
  FILE *fp;
  char *filename;
  char *prevline;
  int errstat;

  enum fstatus status;
};

static unsigned int read_configuration (struct internal *vars, unsigned int openstat, 
                                        char **errstr);
static void reopen_log_file (struct internal *vars, struct finfo *fstrc);
static int open_log_file (struct internal *vars, struct finfo *fstrc);
static void close_log_file (struct finfo *fstrc, enum fstatus status);
static void monitor_log_files (struct internal *vars, int activelogs);
static struct rule *check_entry (struct list_s *rlist, char *entry,
				 char **substr, int *rc);
static int parse_filenames (struct d_array *array,
			    void (fill_array) (struct d_array * logs,
					       char *fname), char *buffer);
static void initialize_logs (struct d_array *logs, char *fname);
static void initialize_rules (struct d_array *rules, char *fname);
static void get_timestamp_str (char *str);
#if 0
static void get_timestamp (struct tm *tmptr);
#endif
static void daemonize ();
static void clear_nuhe (struct internal *vars);
static void exit_cleanly (struct internal *vars);
static void free_rule_data (struct list_s *rlist);
static void free_action_data (struct list_s *alist);
static void free_event_data (struct list_s *elist);
static unsigned int create_pid_file (struct internal *vars);
static void register_signal_handlers (struct internal *vars);
static void d_arrays__create (struct internal *vars);
static void slists__create (struct internal *vars);
#ifdef NODEMGR_SUPPORT
static void run_defunc_state (struct internal *vars);
#endif
static void sigterm_handler (int signo);
static void sighup_handler (int signo);

static FILE *logfp = NULL;
static char *cfptr = NULL;
static enum logmethod logmode = LOG_FILE;
static enum runmethod runmode = STANDALONE;

int
main (int argc, char **argv)
{
  struct stat buf;
  struct internal vars;
  struct finfo *fstrc;
  int ich;
  int trtn;
  int fcnt;
  int fdi;
  char *conferr;

  /* Zero out & initialize internal struct members */
  memset (vars.manageraddr, 0, sizeof (vars.manageraddr));
  memset (vars.sensorname, 0, sizeof (vars.sensorname));
  memset (vars.logfile, 0, sizeof (vars.logfile));
  memset (vars.pidfile, 0, sizeof (vars.pidfile));
  memset (vars.ruledir, 0, sizeof (vars.ruledir));
  memset (vars.eventfile, 0, sizeof (vars.eventfile));
  memset (vars.logfilesfile, 0, sizeof (vars.logfilesfile));
  memset (vars.whitelistfile, 0, sizeof (vars.whitelistfile));
  memset (vars.rulesfile, 0, sizeof (vars.rulesfile));
  memset (vars.blocklistcmd, 0, sizeof (vars.blocklistcmd));
  memset (vars.privatekey, 0, sizeof (vars.privatekey));
  memset (vars.certificate, 0, sizeof (vars.certificate));
  memset (vars.cafile, 0, sizeof (vars.cafile));
  memset (vars.cadir, 0, sizeof (vars.cadir));
  memset (failstore, 0, sizeof (failstore));

  vars.cfile = NULL;
  vars.rules = NULL;
  vars.lmsg = NULL;

  vars.run_stat = DAEMONIC;
  vars.manageralrt = 0;
  vars.active_fmon = 1;
  vars.connretries = 0;

  vars.argvptr = argv;
  vars.errstr = NULL;

  
  fcnt = 0;
  opterr = 0;
  while ((ich = getopt (argc, argv, "fvhrsc:")) != -1)
    switch (ich)
      {
      case 'f':
	vars.run_stat = FOREGROUD;
	break;
      case 'v':
	printf ("Nuhe log monitor version %s\n", VERSION);
	exit (0);
	break;
      case 'h':
	printf
	  ("Nuhe - action capable log file monitor %s\nCopyright (C) 2006-2023 Tuomo Makinen\n\nUsage:\n  -c  <file>  Configuration file for Nuhe log monitor\n\n  -f  Foreground mode\n  -r  Redirect logging to stdout\n  -s  <file1> <file2> ...  Log files to scan\n  -v  Show version of Nuhe log monitor\n  -h  Prints this message\n\n",
	   VERSION);
	exit (0);
	break;
      case 'c':
	cfptr = optarg;
	break;
      case 'r':
	logmode = LOG_STDOUT;
	break;
      case 's':
	vars.run_stat = SCANLOGS;
	logmode = LOG_STDOUT;
	break;
      case '?':
	fprintf (stderr, "Error: unknown option, -%c\n", optopt);
	exit (1);
	break;
      }

  if (vars.run_stat == DAEMONIC && logmode == LOG_STDOUT)
    {
      fprintf (stderr,
	       "Error: you can only specify stdout logging (-s) with foreground mode (-f)\n");
      exit (1);
    }

  /* Create list_s's for rules, actions and events */
  slists__create (&vars);

  /* Create d_array's for logs and rules */
  d_arrays__create (&vars);

  /* Read configuration file */
  if (read_configuration (&vars, 0, &conferr) > 0)
    {
      enum status stat_bu;

      /* Error occured during file opening/parsing 
       * set run_stat to STARTING
       */
      stat_bu = vars.run_stat;
      vars.run_stat = STARTING;

      fprintf (stderr, "%s",conferr);
      fprintf (stderr,"%s",
              "Warning: trying to open configuration file backup\n");
      free (conferr);

      /* Free & destroy allocated data before trying to read backup
       * configuration file
       */
      clear_nuhe (&vars);
      vars.run_stat = stat_bu;

      /* Recreate d_arrays */
      d_arrays__create (&vars);

      /* We try to open and read backup configuration file; this adds 
       * fault tolerance if we are going to connect to node manager
       */
      if (read_configuration (&vars, 1, &conferr) > 0)
        {
          /* No luck ... */
          fprintf (stderr,"%s", conferr);
          fprintf (stderr,
                  "Error: failed to read configuration file for Nuhe, exiting\n");
          free (conferr);
          exit (1);
        }
     }

  /* Specific for DAEMONIC and FOREGROUND states */
  if (vars.run_stat != SCANLOGS)
    {
      /* Check if the pid file exists; we do not want run
       * multiple instances of nuhe
       */
      if ((stat (vars.pidfile, &buf) != -1))
	{
	  fprintf (stderr, "Error: PID file exists (%s)\n", vars.pidfile);
          exit (1);
	}

      /* umask for nuhe files */
      umask (077);

      /* Open logging file. If we are running standalone mode
       * alerts go to the logfile
       */
      if (logmode != LOG_STDOUT)
        {
          logfp = fopen (vars.logfile, "a");
          if (logfp == NULL)
            {
              fprintf (stderr,
                       "Error: can't open or create %s, %s\n",
                       vars.logfile, strerror (errno));
              exit (1);
            }
        }
    }
  /* In SCANLOGS state initialize logs d_array for files given on command line */
  else if (vars.run_stat == SCANLOGS && optind < argc)
    {
      while (optind < argc)
        initialize_logs (vars.logs, argv[optind++]);
    }

  /* Parse rules */
  if (parse_rules (&vars) > 0)
    {
      /* We're standalone sensor; bail out */
      if (runmode != NODEMANAGER)
        {
          fprintf (stderr, "%s",vars.errstr);
          fprintf (stderr, "%s","Error: failed to read rules, exiting\n");
          exit (1);
        }
#ifdef NODEMGR_SUPPORT
      /* Decline to DEFUNC state */
      else
        {
          vars.errstr[strlen(vars.errstr) - 1] = '\0';
          logwriter (vars.errstr);
          logwriter ("Warning: Nuhe started in DEFUNC state");
          run_defunc_state (&vars);
        }
#endif
    }

  /* Open monitored log files */
  for (fdi = 0; fdi < d_array_get_size (vars.logs); fdi++)
    {
      fstrc = (struct finfo *) vars.logs->element[fdi]->payload;
      /* no more files to monitor */
      if (fstrc == NULL)
	break;

      if (open_log_file (&vars, fstrc) == -1)
	{
	  fprintf (stderr, "Skipping file: %s, %s\n", fstrc->filename, strerror (errno));
	  fstrc->status = DISABLED;
	  fstrc->errstat = ERRSTATMAX;
	}
      else
	{
	  fprintf (stderr, "Registering file: %s \n", fstrc->filename);
	  fstrc->status = ACTIVE;
	  fstrc->errstat = 0;
	  fcnt++;
	}

      fstrc->prevline = NULL;
    }

  /* No files to monitor; disable file monitor */
  if (!fcnt)
    vars.active_fmon = 0;

  /* If sensor is standalone and file monitor is disabled exit */
  if (vars.active_fmon < 1 && runmode != NODEMANAGER)
    {
      fprintf (stderr, "Error: zero files to monitor, exiting\n");
      exit (1);
    }

  /* In SCANLOGS state we exit after given logs are analyzed */
  if (vars.run_stat == SCANLOGS)
    monitor_log_files (&vars, fcnt);

  /* Rest of the DAEMONIC and FOREGROUND specific initializations */
  else
    {
      if (vars.run_stat == DAEMONIC)
        {
          /* Daemonize */
          printf ("Forking to background\n");
          daemonize ();
        }
 
      register_signal_handlers (&vars);

      /* We're running */
      quit_flag = RUNNING;

      /* Log Nuhe startup */
      if (vars.run_stat == DAEMONIC)
        logwriter ("Nuhe daemon started");
      else
        logwriter ("Nuhe started in foreground mode");

      /* Start event handler thread */
      if ((trtn =
          pthread_create (&vars.event_t, NULL, (void *) monitor_events, &vars)))
        {
          fprintf (stderr, "Error: pthread_create, %s\n", strerror (trtn));
          exit (1);
        }

#ifdef NODEMGR_SUPPORT
      if (runmode == NODEMANAGER)
        {
          /* Start node manager connector thread */
          if ((trtn =
	      pthread_create (&vars.nodem_t, NULL, (void *) nuhe_connector,
			      &vars)))
	    {
	      fprintf (stderr, "Error: pthread_create, %s\n", strerror (trtn));
	      exit (1);
	    }    
        }
#endif

      create_pid_file (&vars);

#ifdef NODEMGR_SUPPORT
      /* We don't have any files to monitor but we're running
       * in NODEMANAGER mode; main thread waits event handler
       * and node manager connector threads to exit
       */
      if (vars.active_fmon < 1 && runmode == NODEMANAGER)
        {
          logwriter ("Warning: zero files to monitor; file monitor disabled");
          exit_cleanly (&vars);
        }
#endif

      /* Monitor log files forever */
      monitor_log_files (&vars, fcnt);
    }

  return 0;
}

/* daemonizer */
static void
daemonize ()
{
  pid_t pid;

  pid = fork ();
  if (pid < 0)
    {
      perror ("fork() failed");
      exit (errno);
    }
  else if (pid != 0)
    exit (0);
  else
    {
      setsid ();
 
      chdir ("/");

      /* Redirect STDIN, STDOUT and STDERR
       * to /dev/null
       */

      FILE *fp;
      fp = freopen ("/dev/null", "r", stdin);
      fp = freopen ("/dev/null", "w", stdout);
      fp = freopen ("/dev/null", "w", stderr);
   
      /* what a way to get rid off compiler warnings */
      if (fp == NULL){ }
    }
}

static void
register_signal_handlers (struct internal *vars)
{
  signal (SIGTERM, sigterm_handler);
  signal (SIGHUP, sighup_handler);

  /* Initialize SIGINT handler, if running on foreground */
  if (vars->run_stat == FOREGROUD)
    signal (SIGINT, sigterm_handler);
}

static unsigned int
create_pid_file (struct internal *vars)
{
  FILE *pidf;
  
  pidf = fopen (vars->pidfile, "w");
  if (pidf == NULL)
    {
      char *err;

      err = (char *) calloc (STRBUF, sizeof (char));
      if (err == NULL)
        {
          /* perror ("calloc() failed"); */
          exit (1);
        }

      snprintf (err, STRBUF, "Error: can't create %s, %s\n",
                vars->pidfile, strerror (errno));
      logwriter (err);
      free (err);

      return 1;
    }

  fprintf (pidf, "%d\n", (int) getpid ());
  fclose (pidf);

  return 0;    
}

static void
d_arrays__create (struct internal *vars)
{
  /* Create d_array for finfo structures */
  vars->logs = d_array_create ();
  if (vars->logs == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  /* Create d_array for rulefiles */
  vars->rules = d_array_create ();
  if (vars->rules == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }
}

static void
slists__create (struct internal *vars)
{
  /* Create list_s for rules */
  vars->rlist = slist_create ();
  if (vars->rlist == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  /* Create list_s for actions */
  vars->alist = slist_create ();
  if (vars->alist == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  /* Create list_s for events */
  vars->elist = slist_create ();
  if (vars->alist == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }
}

#ifdef NODEMGR_SUPPORT
/* Here we switch sensor state to DEFUNC; this state is only useful
 * when sensor is configured to connect node manager. DEFUNC state
 * means that there is some error in sensor rule or configuration
 * files and this state is used to allow updating of rule and 
 * configuration files and restarting or quiting sensor
 */
static void
run_defunc_state (struct internal *vars)
{
  if (vars->run_stat == DAEMONIC)
    daemonize ();

  register_signal_handlers (vars);
  create_pid_file (vars);

  /* Switch to DEFUNC state */
  vars->run_stat = DEFUNC;

  /* Start only node manager connector */
  nuhe_connector (vars);

  exit_cleanly (vars);
}
#endif

/* This function parse multiple filenames from configuration
 * file and calls fill_array function pointer to add it to d_array 
 */
static int
parse_filenames (struct d_array *array,
		 void (fill_array) (struct d_array * logs, char *fname),
		 char *buffer)
{
  char *cptr;
  char *lfile;
  char fname[STRBUF] = {
    0
  };

  cptr = &buffer[0];
  lfile = &fname[0];

  while (isspace ((int) *cptr))
    cptr++;
  while (isgraph ((int) *cptr) || *cptr == '/')
    *lfile++ = *cptr++;
  while (isspace ((int) *cptr))
    cptr++;

  (*fill_array) (array, fname);

  if (*cptr == '\\')
    return 0;
  else
    return 1;
}

/* Initializes logs d_array with finfo structs */
static void
initialize_logs (struct d_array *logs, char *fname)
{
  int index;

  index = d_array_add_element (logs);
  if (index <= -1)
    {
      perror ("d_array_add_element () failed");
      exit (1);
    }

  logs->element[index]->payload = (void *) malloc (sizeof (struct finfo));
  if (logs->element[index]->payload == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  ((struct finfo *) logs->element[index]->payload)->fp = NULL;
  ((struct finfo *) logs->element[index]->payload)->status = INACTIVE;
  ((struct finfo *) logs->element[index]->payload)->filename =
    (void *) calloc (strlen (fname) + 1, sizeof (char));
  if (((struct finfo *) logs->element[index]->payload)->filename == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  strcpy (((struct finfo *) logs->element[index]->payload)->filename, fname);
}

/* Initializes rules d_array with rulefile structs */
static void
initialize_rules (struct d_array *rules, char *fname)
{
  int index;
  int i;
  char *fnptr;

  i = strlen (fname);
  fnptr = &fname[i];
  for (; i > 0; i--)
    {
      fnptr = &fname[i];
      if (*fnptr == '/')
	{
	  fnptr = &fname[i + 1];
	  break;
	}
      fnptr--;
    }

  index = d_array_add_element (rules);
  if (index <= -1)
    {
      perror ("d_array_add_element () failed");
      exit (1);
    }

  rules->element[index]->payload =
    (void *) malloc (sizeof (struct rulefiles));
  if (rules->element[index]->payload == NULL)
    {
      perror ("malloc() failed");
      exit (1);
    }

  ((struct rulefiles *) rules->element[index]->payload)->fpath =
    (void *) calloc (strlen (fname) + 1, sizeof (char));
  if (((struct rulefiles *) rules->element[index]->payload)->fpath == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }
  strcpy (((struct rulefiles *) rules->element[index]->payload)->fpath,
	  fname);

  ((struct rulefiles *) rules->element[index]->payload)->fname =
    (void *) calloc (strlen (fnptr) + 1, sizeof (char));
  if (((struct rulefiles *) rules->element[index]->payload)->fname == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }
  strcpy (((struct rulefiles *) rules->element[index]->payload)->fname,
	  fnptr);
}

static unsigned int
read_configuration (struct internal *vars, unsigned int openstat, char **errstr)
{
  FILE *fp;
  char buffer[BUFSIZE];
  char *cptr;
  int rlogs;
  int smode;
  int i;
  int use_managed_logfiles = 0;
  char cnfile[STRBUF] = { 0 };

  struct d_array *dptr = NULL;
  void (*fill_array) (struct d_array * logs, char *fname) = NULL;

  char *confiles[] =
    {
      "../etc/nuhed/nuhed.conf",
      "/usr/local/etc/nuhed.conf",
      "/usr/local/etc/nuhed/nuhed.conf",
      "/etc/nuhed/nuhed.conf", "/etc/nuhed.conf",
      "./nuhed.conf"
    };

  rlogs = 1;
  smode = 1;

  /* Allocate space for error message */
  *errstr =  (char *) calloc (BUFSIZE, sizeof (char));
  if (*errstr == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }

  /* Search for configuration file */
  if (cfptr == NULL)
    {
      for (i = 0; i < 6; i++)
	{
          memset (cnfile, 0, sizeof (cnfile));

          openstat ? snprintf (cnfile, STRBUF, "%s.%s", confiles[i], "backup") :
                     snprintf (cnfile, STRBUF, "%s", confiles[i]);

	  fp = fopen (cnfile, "r");
	  if (fp != NULL)
	    break;
	}
    }
  /* Configuration file given from command line */
  else
    {
      openstat ? snprintf (cnfile, STRBUF, "%s.%s", cfptr, "backup") : 
                 snprintf (cnfile, STRBUF, "%s", cfptr);

      fp = fopen (cnfile, "r");
    }

  if (fp == NULL)
    {
      snprintf (*errstr, BUFSIZE,
	       "Error: can't open configuration file %s, %s\n",
	       cnfile, strerror (errno));
      return 1;
    }

  /* vars->cfile contains configuration file name */
  vars->cfile = (char *)  calloc (strlen (cnfile) + 1, sizeof (char));
  if (vars->cfile == NULL)
    {
      perror ("calloc() failed");
      exit (1);
    }
  memcpy (vars->cfile, cnfile, strlen (cnfile));

  printf ("Using %s configuration file\n", vars->cfile);

  i = 0;
  while (fgets (buffer, sizeof (buffer), fp) != NULL)
    {

      if (!strncmp (buffer, "#", 1))
	continue;
      else if (*buffer == '\0')
	continue;
      else if (rlogs == 0)
	{
	  if (vars->run_stat == SCANLOGS && dptr == vars->logs)
	    break;
	  else
	    rlogs = parse_filenames (dptr, fill_array, buffer);
	}
      else if (!strncmp (buffer, "Rules", 5))
	{
	  fill_array = &initialize_rules;
	  dptr = vars->rules;
	  rlogs = -1;
	}
      else if (vars->run_stat != SCANLOGS && !strncmp (buffer, "Logs", 4))
	{
	  fill_array = &initialize_logs;
	  dptr = vars->logs;
	  rlogs = -1;
	}
      else if (!strncmp (buffer, "lastmsg", 7))
        {
          int i;
          int erroffset;
          const char *error;

          cptr = &buffer[7];
          while (isspace ((int) *cptr))
            cptr++;
          if (*cptr != '\"')
            {
              snprintf (*errstr, BUFSIZE, "Error reading configuration file: invalid value for lastmsg\n");
              return 1;
            }
          cptr++;

          i = strlen (cptr) - 1;
          while (i > 0)
            {
              if (cptr[i] == '\"')
                {
                  cptr[i] = '\0';
                  break;
                }
              if (!isspace (cptr[i]))
                {
                  snprintf (*errstr, BUFSIZE, "Error reading configuration file: invalid value for lastmsg\n");
                  return 1;
                }
              i--;
            }

          vars->lmsg = pcre_compile (cptr, 0, &error, &erroffset, NULL);
          if (vars->lmsg == NULL)
            {
              snprintf (*errstr, BUFSIZE,
                       "Error reading configuration file: failed to compile regular expression for %s lastmsg string\n",
                       cptr);
              return 1;
            }
        }
      else if (!strncmp (buffer, "Standalone", 10) && 
	       vars->run_stat != SCANLOGS)
	{
	  sscanf (buffer, "Standalone %d", &smode);
          if (smode < 0 || smode > 1)
            {
              snprintf(*errstr, BUFSIZE,
                       "Error reading configuration file: Standalone value must be 0 or 1\n");
              return 1;

            }
	  if (!smode)
	    {
#ifdef NODEMGR_SUPPORT
	      runmode = NODEMANAGER;
#else
snprintf (*errstr, BUFSIZE,
	"You must recompile Nuhe to use node manager; ./configure --enable-nodemgr\n");
return 1;
#endif
	    }
	}
      else if (!strncmp (buffer, "Pidfile", 7))
	sscanf (buffer, "Pidfile %255s", vars->pidfile);
      else if (!strncmp (buffer, "Logfile", 7))
	sscanf (buffer, "Logfile %255s", vars->logfile);
      else if (!strncmp (buffer, "Ruledir", 7))
        sscanf (buffer, "Ruledir %255s", vars->ruledir);
      else if (!strncmp (buffer, "Eventfile", 9))
        sscanf (buffer, "Eventfile %255s", vars->eventfile);
      else if (!strncmp (buffer, "Blocklist", 9))
        sscanf (buffer, "Blocklist %[^\n]", vars->blocklistcmd);
      else if (!strncmp (buffer, "Actrulesfile", 9))
        sscanf (buffer, "Actrulesfile %255s", vars->rulesfile);
      else if (!strncmp (buffer, "FailsafeStore", 13))
        sscanf (buffer, "FailsafeStore %255s", failstore);
      else if (!strncmp (buffer, "Privatekey", 10))
        sscanf (buffer, "Privatekey %255s", vars->privatekey);
      else if (!strncmp (buffer, "Certificate", 11))
        sscanf (buffer, "Certificate %255s", vars->certificate);
      else if (!strncmp (buffer, "CAfile", 6))
        sscanf (buffer, "CAfile %255s", vars->cafile);
      else if (!strncmp (buffer, "Logdestination", 14))
        {
          sscanf (buffer, "Logdestination %d", &smode);
	  if (logmode != LOG_STDOUT)
	    {
	      if (smode == 1)
	        logmode = LOG_FILE;
	      else if (smode == 2)
	        logmode = LOG_NODEMGR;
	      else if (smode == 3)
                logmode = LOG_BOTH;
	      else
		{
		  snprintf (*errstr, BUFSIZE,
		            "Error reading configuration file: Logdestination value must be 1, 2 or 3\n");
		  return 1;
		}
	    }
	}
      else if (!strncmp (buffer, "NodemanagerAddr", 15))
	sscanf (buffer, "NodemanagerAddr %255s", vars->manageraddr);
      else if (!strncmp (buffer, "NodemanagerPort", 15))
	sscanf (buffer, "NodemanagerPort %u", &vars->managerport);
      else if (!strncmp (buffer, "ConnectionRetries", 17))
	sscanf (buffer, "ConnectionRetries %u", &vars->connretries);
      else if (!strncmp (buffer, "NodeName", 8))
	sscanf (buffer, "NodeName %255s", vars->sensorname);
      else if (!strncmp (buffer, "NodemanagerAlerts", 17))
        sscanf (buffer, "NodemanagerAlerts %u", &vars->manageralrt);
      else if (!strncmp (buffer, "WhitelistFile", 13))
        sscanf (buffer, "WhitelistFile %255s", vars->whitelistfile);
        
      if (!strncmp (buffer, "LogfilesFile", 12))
        sscanf (buffer, "LogfilesFile %255s", vars->logfilesfile);
      /* read first string from multiple lines (Rules and Logs) */
      if (rlogs == -1)
	{
	  if (!strncmp (buffer, "Logs", 4)){
            FILE *fpl;
            char mylogfile[512] = {0};
            fpl  = fopen (vars->logfilesfile, "r");
            if (fpl != NULL){
	      use_managed_logfiles = 1;
              while (fgets (mylogfile, sizeof (mylogfile), fpl) != NULL ){
		mylogfile[strlen(mylogfile) - 1] = 0;
		if (mylogfile[0] != '\0' && !isspace(mylogfile[0]) && strstr(mylogfile, "/etc") == NULL){
                  rlogs = parse_filenames (dptr, fill_array, mylogfile);
		}
              }
              fclose (fpl);
	    }
	  }

	  if (use_managed_logfiles == 0){
            cptr = &buffer[0];
	    while (!isspace ((int) *cptr))
	      cptr++;

	    cptr++;
	    rlogs = parse_filenames (dptr, fill_array, cptr);
          }
       }
    }
  fclose (fp);

  /* Do some validation for configuration parameters */
  if (vars->rules == NULL)
    {
      snprintf (*errstr, BUFSIZE,
	       "Error reading configuration file: no rule files specified with Rule\n");
      return 1;
    }
  if (*vars->pidfile == '\0')
    {
      memcpy (vars->pidfile, "/var/run/nuhed.pid", STRBUF);
    }
  if (*vars->logfile == '\0')
    {
      snprintf (*errstr, BUFSIZE,
	       "Error reading configuration file: You must specify logfile\n");
      return 1;
    }
  if (*vars->eventfile == '\0')
    {
      snprintf (*errstr, BUFSIZE,
               "Error reading configuration file: You must specify eventfile\n");
      return 1;
    }
  if (*vars->logfilesfile == '\0')
    {
      snprintf (*errstr, BUFSIZE,
               "Error reading configuration file: You must specify Logfilesfile\n");
      return 1;
    }

  if (*vars->rulesfile == '\0')
    {
      snprintf (*errstr, BUFSIZE,
               "Error reading configuration file: You must specify rulesfile\n");
    }

#ifdef NODEMGR_SUPPORT
  if (runmode == NODEMANAGER && *vars->manageraddr == '\0')
    {
      snprintf (*errstr, BUFSIZE,
	       "Error reading configuration file: no node manager address specified with NodemanagerAddr\n");
      return 1;
    }
  if (runmode == NODEMANAGER && *vars->privatekey == '\0')
    {
      snprintf (*errstr, BUFSIZE,
               "Error reading configuration file: no privatekey specified with Privatekey\n");
      return 1;
    }
  if (runmode == NODEMANAGER && *vars->certificate == '\0')
    {
      snprintf (*errstr, BUFSIZE,
               "Error reading configuration file: no certificate specified with Certificate\n");
      return 1;
    }
  if (runmode == NODEMANAGER && *vars->cafile == '\0')
    {
      snprintf (*errstr, BUFSIZE,
               "Error reading configuration file: no trusted CA file specified with CAfile\n");
      return 1;
    }
  if (runmode == NODEMANAGER && *vars->ruledir == '\0')
    {
      snprintf (*errstr, BUFSIZE,
               "Error reading configuration file: no directory containing rulesets specified with Ruledir\n");
      return 1;
    }
  if (runmode == NODEMANAGER && *failstore == '\0')
    {
      snprintf (*errstr, BUFSIZE,
               "Error reading configuration file: no filename specified with FailsafeStore\n");
      return 1;
    }
#endif
  if (runmode == STANDALONE)
    {
      if (logmode == LOG_NODEMGR || logmode == LOG_BOTH)
        {
          snprintf (*errstr, BUFSIZE,
	           "Error reading configuration file: you must specify Logdestination to 1 (local) for standalone sensor\n");
	  return 1;
	}
    }
  if (vars->run_stat != SCANLOGS && rlogs != 1)
    {
      snprintf (*errstr, BUFSIZE,
	       "Error reading configuration file: no monitored files specified with Logs\n");
      return 1;
    }
  if (*vars->sensorname == '\0')
    {
      /* Use system name for sensorname */
      struct utsname buf;
      if (uname (&buf) == -1)
	{
	  snprintf (*errstr, BUFSIZE,
		   "Error reading configuration file: no NodeName specified and uname() failed, %s\n",
		   strerror (errno));
	  return 1;
	}
      memcpy (vars->sensorname, buf.nodename, STRBUF);
    }

  /* All is ok, free errstr */
  free (*errstr);
  *errstr = NULL;

  return 0;
}

/* Handle log file opening */
static int
open_log_file (struct internal *vars, struct finfo *fstrc)
{
  fstrc->fp = fopen (fstrc->filename, "r");

  if (fstrc->fp == NULL)
    return -1;
  else
    {
      /* seek to end of the file, if file is
       * opened during daemon startup.
       */
      if (vars->run_stat != SCANLOGS && fstrc->status == INACTIVE)
	{
	  if (fseek (fstrc->fp, 0, SEEK_END) < 0)
	    {
	      fclose (fstrc->fp);
	      fstrc->fp = NULL;
	      return -1;
	    }
	}
      return 1;
    }
}

/* Close fd, free memory allocated for previous line read from log
 * and set status for monitored file
 */
static void
close_log_file (struct finfo *fstrc, enum fstatus status)
{
  if (fstrc->fp != NULL)
    {
      fclose (fstrc->fp);
      fstrc->fp = NULL;
    }

  /* Make sure fstrc->prevline is freed */
  if (fstrc->prevline != NULL)
    {
      free (fstrc->prevline);
      fstrc->prevline = NULL;
    }

  fstrc->status = status;
}

/* This function tries to reopen log file and update it's status */
static void
reopen_log_file (struct internal *vars, struct finfo *fstrc)
{
  char alarmstr[LOGBUF];
  int rval;

  /* Make sure fstrc->prevline is freed */
  if (fstrc->prevline != NULL)
    {
      free (fstrc->prevline);
      fstrc->prevline = NULL;
    }

  rval = open_log_file (vars, fstrc);

  if (rval == -1)
    {
      snprintf (alarmstr, LOGBUF,
		"[ERROR] %s, can't reopen log file %s, %s",
		vars->sensorname, fstrc->filename, strerror (errno));
      send_alarm ((const char *) alarmstr);
      fstrc->errstat--;
    }
  else
    {
      snprintf (alarmstr, LOGBUF,
		"[INFO] %s, file %s reopened",
		vars->sensorname, fstrc->filename);
      send_alarm ((const char *) alarmstr);
      fstrc->status = ACTIVE;
      fstrc->errstat = 0;
    }
}

/* Main loop for Nuhe: monitor logfiles, check files against rules
 * and generate alerts
 */
static void
monitor_log_files (struct internal *vars, int activelogs)
{
  char buffer[LOGBUF];
  char alarmstr[LOGBUF];
  struct stat sbuf, fbuf;
  struct timeval timeout;
  struct finfo *fstrc;
  long pos;
  int i;

  while (1)
    {
      /* No more files to monitor ... */
      if (!activelogs)
	{
          if (runmode != NODEMANAGER)
            {
              snprintf (alarmstr, LOGBUF,
                        "[ERROR] %s, no files to monitor, exiting",
                        vars->sensorname);
              send_alarm ((const char *) alarmstr);
              quit_flag = QUIT;
            }
#ifdef NODEMGR_SUPPORT
          else
            {
              snprintf (alarmstr, LOGBUF,
                        "[ERROR] %s, no files to monitor; file monitor disabled",
                         vars->sensorname);
              send_alarm ((const char *) alarmstr);
              vars->active_fmon = 0;
              exit_cleanly (vars);
            }
#endif
	}

      for (i = 0; i < d_array_get_size (vars->logs); i++)
	{
	  fstrc = (struct finfo *) vars->logs->element[i]->payload;

	  /* Workaround for -s command line option */
	  if (vars->run_stat == SCANLOGS)
	    {
	      if (fstrc->status != DISABLED)
		{
		  printf ("*** Analyzing file %s ***\n\n", fstrc->filename);
		  while (1)
		    {
		      if (fgets (buffer, sizeof (buffer), fstrc->fp) == NULL)
			break;
		      else
			process_alert_data (vars, fstrc->filename, buffer);
		    }
		}
	      if (i == d_array_get_size (vars->logs) - 1)
		{
		  exit_cleanly (vars);
		}
	      continue;
	    }

	  timeout.tv_sec = 0;
	  timeout.tv_usec = FMON_SLEEPTIME;

#ifdef NODEMGR_SUPPORT
          /* FREEZE state means that only node manager connector thread should be 
           * active; wait for state to change
           */
          if (vars->run_stat == FREEZE)
            {
              struct finfo *cf;
              int i;

              pthread_mutex_lock (&freeze_sync.state_mutex);
              while (vars->run_stat == FREEZE)
	        pthread_cond_wait (&freeze_sync.state_cv,
                                   &freeze_sync.state_mutex);
	      pthread_mutex_unlock (&freeze_sync.state_mutex);

              /* Back from FREEZE state, SEEK_END all monitored files to
               * prevent alert generation from content that has come out
               * to monitored files during FREEZE state. Not sure if this
               * is agreeable behaviour for all cases ?
               */
              for (i = 0; i < d_array_get_size (vars->logs); i++)
                {
                  cf = (struct finfo *) vars->logs->element[i]->payload;
                  if (cf->fp != NULL)
                    {
                      fseek (cf->fp, 0, SEEK_END);
                      clearerr (cf->fp);
                    }
                }
            }
#endif

	  /* If quit_flag is set to >= 1 we have received SIGTERM, SIGINT or SIGHUP
	   * and should exit cleanly 
	   */
	  if (quit_flag != RUNNING)
	    exit_cleanly (vars);
          
	  /* Break from loop if uninitialized element is reached */
	  if (fstrc == NULL)
	    break;

	  /* Just jump away from disabled or inactive files */
	  if (fstrc->status == DISABLED || fstrc->status == INACTIVE)
	    continue;
	  else if (fstrc->status == REOPEN)
	    {
	      /* File is marked to be reopened... this indicates that file
	       * has possibly rotated; try to reopen it
	       */
	      if (fstrc->errstat == ERRSTATMAX)
		{
		  /* file will be disabled; too many errors */
		  snprintf (alarmstr, LOGBUF,
			    "[ERROR] %s, too many errors with %s file, disabling it from monitoring",
			    vars->sensorname, fstrc->filename);
		  send_alarm ((const char *) alarmstr);
		  close_log_file (fstrc, DISABLED);
		  activelogs--;
		}
	      else
		{
		  /* sleep for a while, before reopening file */
		  timeout.tv_sec = 1;
		  timeout.tv_usec = 0;
		  select (0, NULL, NULL, NULL, &timeout);
		  reopen_log_file (vars, fstrc);
		}

	      continue;
	    }

	  /* Do some validation checks for monitored files */
	  fstat (fileno (fstrc->fp), &fbuf);
	  if (fbuf.st_nlink == 0)
	    {
	      snprintf (alarmstr, LOGBUF,
			"[ALERT] %s, someone removed %s file (probably log is rotated)\
, hard link count reached zero, trying to reopen file...",
			vars->sensorname, fstrc->filename);
	      send_alarm ((const char *) alarmstr);
	      close_log_file (fstrc, REOPEN);
	      continue;
	    }

	  if (fbuf.st_nlink > 0 && stat (fstrc->filename, &sbuf) != -1)
	    {
	      if (sbuf.st_ino != fbuf.st_ino)
		{
		  snprintf (alarmstr, LOGBUF,
			    "[ALERT] %s, filename %s inode has changed.\
 File exists with different name and someone has created another file with %s name, trying to reopen file...", vars->sensorname, fstrc->filename, fstrc->filename);
		  send_alarm ((const char *) alarmstr);
		  close_log_file (fstrc, REOPEN);
		  continue;
		}
	    }
	  else
	    {
	      snprintf (alarmstr, LOGBUF,
			"[ALERT] %s, file %s has positive hard link count, \
but it's filename does not exists. this probably means that someone has renamed file, trying to reopen file...", vars->sensorname, fstrc->filename);
	      send_alarm ((const char *) alarmstr);
	      close_log_file (fstrc, REOPEN);
	      continue;
	    }

	  /* Read new line from log file and process data */
	  pos = ftell (fstrc->fp);
	  if (fgets (buffer, sizeof (buffer), fstrc->fp) == NULL)
	    {
	      fseek (fstrc->fp, pos, SEEK_SET);
	      clearerr (fstrc->fp);
	      select (0, NULL, NULL, NULL, &timeout);
	    }
	  else
            {
              /* Check for syslog "last message repeated n times" string.
               * if found, send previous string as many times as it repeated
               */
              if (vars->lmsg != NULL)
                {
                  int rc;
                  int i;
                  char *substr[SUBSTRMAX];

                  if (fstrc->prevline != NULL)
                    {
                      match_pcre (vars->lmsg, buffer, substr, &rc);
                      if (rc > 0)
                        {
                          for (i = 0; i < atoi (substr[1]); i++)
                            process_alert_data (vars, fstrc->filename, fstrc->prevline);
                        }
                      free (fstrc->prevline);
                      fstrc->prevline = NULL;
                    }
                  fstrc->prevline = (char *) calloc (strlen (buffer) + 1, sizeof (char));
                  if (fstrc->prevline == NULL)
                    logwriter ("CRITICAL: can't allocate memory for previously read line, calloc()");
                  else
                    memcpy (fstrc->prevline, buffer, strlen (buffer));
                }

              /* Process line read from log file */
	      process_alert_data (vars, fstrc->filename, buffer);
            }
	}
    }
}

/* Returns pointer to appropriate rule structure, if Pcre match is succesful */
static struct rule *
check_entry (struct list_s *rlist, char *entry, char **substr, int *rc)
{
  struct node *current;
  struct rule *payload;

  *rc = -1;
  current = rlist->head;

  while (current != NULL)
    {
      payload = (struct rule *) current->payload;
      match_pcre (payload->re, entry, substr, rc);

      if (*rc > 0)
        return payload;

      current = current->next;
    }
  return NULL;
}

static void
sigterm_handler (int signo)
{
  quit_flag = QUIT;
  
  if (signo == SIGTERM)
    logwriter ("Received the SIGTERM signal - exiting");
  else if (signo == SIGINT)
    logwriter ("Received the SIGINT signal - exiting");
}

static void
sighup_handler (int signo)
{
  quit_flag = RESTART;

  logwriter ("Received the SIGHUP signal - restarting...");
}

static void
exit_cleanly (struct internal *vars)
{
  /* Good housekeeping */  
  clear_nuhe (vars);

  /* Nuhe needs to be restarted */
  if (quit_flag == RESTART)
    {
      if (execvp ((char * ) (vars)->argvptr[0],
                  (char ** ) (vars)->argvptr) == -1)
        {
          char err[128] = { 0 };

          snprintf(err, 128, "Error: can't restart Nuhe, execvp() failed: %s",
                   strerror (errno));
          logwriter (err);
        }
    }

  if (logmode != LOG_STDOUT)
    fclose (logfp);

  exit (0);
}

static void
clear_nuhe (struct internal *vars)
{
  struct finfo *fstrc;
  struct rulefiles *rstrc;
  struct d_array *logs;
  struct d_array *rules;
  int darr_sz;
  int i;

  logs = vars->logs;
  rules = vars->rules;

  /* In DEFUNC or STARTING state we are only running main thread */
  if (vars->run_stat != DEFUNC &&
      vars->run_stat != STARTING)
    {
      /* If we are running with file monitoring disabled, we
       * block at pthread_join (); other threads run until
       * quit_flag is set to !RUNNING 
       */

      /* Wait event handler thread to exit */
      //if (vars->run_stat != SCANLOGS)
      //  pthread_join (vars->event_t, NULL);
      pthread_kill(vars->event_t, SIGTERM);
#ifdef NODEMGR_SUPPORT
      /* Wait node manager connector thread to exit */
      //pthread_join (vars->nodem_t, NULL);
      if (runmode == NODEMANAGER)
        pthread_kill (vars->nodem_t, SIGTERM);
#endif
    }  

  if (vars->run_stat != STARTING)
    {
      /* Free rule, action and event structs */
      free_rule_data (vars->rlist);
      free_action_data (vars->alist);
      free_event_data (vars->elist);

      /* Delete rule, action and event lists */
      slist_delete (vars->rlist);
      slist_delete (vars->alist);
      slist_delete (vars->elist);
    }

  darr_sz = d_array_get_size (logs);
  /* Close active fd's and free memory that is allocated
   * for finfo structures
   */
  for (i = 0; i < darr_sz; i++)
    {
      fstrc = (struct finfo *) logs->element[i]->payload;
      if (vars->run_stat != DEFUNC &&
          vars->run_stat != STARTING)
        {
          close_log_file (fstrc, DISABLED);
        }
      free (fstrc->filename);
    }

  /* Delete logs d_array */
  d_array_delete (logs);

  darr_sz = d_array_get_size (rules);
  /* Free rulefile structure memory */
  for (i = 0; i < darr_sz; i++)
    {
      rstrc = (struct rulefiles *) rules->element[i]->payload;
      free (rstrc->fname);
      free (rstrc->fpath);
    }

  /* Delete rules d_array */
  d_array_delete (rules);

  if (vars->cfile != NULL)
    free (vars->cfile);

  if (vars->lmsg != NULL)
    free (vars->lmsg);

  if (vars->errstr != NULL)
    free (vars->errstr);

  if (vars->run_stat != SCANLOGS &&
      vars->run_stat != STARTING)
    {
      if (unlink (vars->pidfile) != 0)
	logwriter ("Error: can't unlink PID file");
    }
}

static void
free_rule_data (struct list_s *rlist)
{
  struct node *current;
  struct node *next;
  struct rule *cell;

  current = rlist->head;

  while (current != NULL)
    {
      next = current->next;
      cell = (struct rule *) current->payload;
      if (cell->actname != NULL)
	free (cell->actname);

      free (cell->desc);
      free (cell->re);
      current = next;
    }
}

static void
free_action_data (struct list_s *alist)
{
  struct node *current;
  struct node *next;
  struct action *cell;
  unsigned int darr_sz;
  int i;

  current = alist->head;

  while (current != NULL)
    {
      next = current->next;
      cell = (struct action *) current->payload;

      /* Delete ignore d_array */
      if (cell->ignore != NULL)
	d_array_delete (cell->ignore);

      if (cell->run_action != NULL)
	{
	  /* Free action from do_action struct */
	  darr_sz = d_array_get_size (cell->run_action);
	  for (i = 0; i < darr_sz; i++)
	    free (((struct do_action *) cell->run_action->element[i]->
		   payload)->action);

	  /* Free d_array */
	  d_array_delete (cell->run_action);
	}
      if (cell->reset != NULL)
        {
          /* Free d_array */
          d_array_delete (cell->reset);
        }

      free (cell->name);
      current = next;
    }
}

static void
free_event_data (struct list_s *elist)
{
  struct node *current;
  struct node *next;
  struct event *cell;
  int i;

  current = elist->head;

  while (current != NULL)
    {
      next = current->next;
      cell = (struct event *) current->payload;

      for (i = 0; i < cell->id_size; i++)
        free (cell->event_id[i]);

      free (cell->event_id);
      free (cell->alertsrc);

      current = next;
    }
}

/* Supports only 24h clock */
static void
get_timestamp_str (char *str)
{
  struct tm *tmptr;
  time_t tm;

  tm = time (NULL);
  tmptr = (struct tm *) localtime (&tm);

  /*strftime (str, STRBUF, "%b %d %H:%M:%S %Y", tmptr);*/
  strftime (str, STRBUF, "%Y-%m-%d %H:%M:%S", tmptr);
}

#if 0
static void
get_timestamp (struct tm *tmptr)
{
  time_t tm;

  tm = time (NULL);
  tmptr = (struct tm *) localtime (&tm);
}
#endif

char *
generate_alarm (struct rule *matched, char *src, char *sensorname, char *data)
{
  char loglevel[6] = {
    0
  };
  char *alarmstr;

  if (matched->aptr->loglevel == CRIT)
    memcpy (loglevel, "CRIT", 4);
  else if (matched->aptr->loglevel == ALERT)
    memcpy (loglevel, "ALERT", 5);
  else if (matched->aptr->loglevel == INFO)
    memcpy (loglevel, "INFO", 4);

  if (matched->aptr->loglevel != NONE)
    {
      alarmstr = (char *) calloc (LOGBUF, sizeof (char));
      if (alarmstr == NULL)
	logwriter ("CRITICAL: can't allocate memory for alarm string, calloc()");
      else
	{
	  if (data[strlen (data) - 1] == '\n')
	    data[strlen (data) - 1] = '\0';

	  snprintf (alarmstr, LOGBUF,
		    "[%s] %s, rule #%d: %s, received from %s. Matched content: %s",
		    loglevel, sensorname, matched->rulenr,
		    matched->desc, src, data);
	}
    }
  else
    alarmstr = NULL;

  return alarmstr;
}

void
process_alert_data (struct internal *vars, char *alertsrc, char *entry)
{
  struct rule *matched = NULL;
  char *substr[SUBSTRMAX];
  char *alarmstr;
  int rc;
  int rc_cnt;

  matched = check_entry (vars->rlist, entry, substr, &rc);
  if (matched != NULL)
    {
      /* PCRE matched */
      if (matched->aptr->loglevel != NONE &&
          (matched->aptr->ltvalue == ALWAYS ||
           vars->run_stat == SCANLOGS))
        {
          /* Generate alarm */
          alarmstr =
            generate_alarm (matched, alertsrc, vars->sensorname, entry);

          if (alarmstr != NULL)
            {
              send_alarm ((const char *) alarmstr);
              free (alarmstr);
            }
        }

      /* process action if specified */
      if (vars->run_stat != SCANLOGS && matched->aptr->id != EMPTY)
        {
          /* Actions are handled by event handler */
          send_event (vars, matched, alertsrc, substr, rc);
        }

      /* Free substrings */
      for (rc_cnt = 0; rc_cnt < rc; rc_cnt++)
        {
          if (substr[rc_cnt] != NULL)
            free (substr[rc_cnt]);
        }
    }
  else
    {
      /* Check if string matches to action rststrings and if we
       * have corresponding event registered then reset it
       */
      search_and_reset_event (vars, entry);
    }
}

void
match_pcre (pcre *re, char *entry, char **substr, int *rc)
{
  int ovector[OVECCOUNT];
  int i;

  *rc = pcre_exec (re,
                   NULL,
                   entry, (int) strlen (entry), 0, 0, ovector, OVECCOUNT);

  /* Collect substrings */
  if (*rc > 0)
    {
      for (i = 0; i < *rc; i++)
        {
          char *substring_start = entry + ovector[2 * i];
          int substring_length = ovector[2 * i + 1] - ovector[2 * i];
          substr[i] =
            (char *) calloc ((substring_length + 1), sizeof (char));
          if (substr[i] == NULL)
            {
              logwriter ("CRITICAL: can't allocate memory for matched substring, calloc()");
              for (i = 0; i < *rc; i++)
                {
                  if (substr[i] != NULL)
                    free (substr[i]);
                }
              *rc = 0;
              break;
            }
          else
            memcpy (substr[i], substring_start, substring_length);
        }
    }
}

void
send_alarm (const char *alarmstr)
{
  if (logmode == LOG_FILE || logmode == LOG_STDOUT)
    {
      /* Just log to normal log file */
      logwriter (alarmstr);
    }
#ifdef NODEMGR_SUPPORT
  else
    {
      char str[STRBUF];
      char *message;

      /* Log alert also on local log file */
      if (logmode == LOG_BOTH)
        logwriter (alarmstr);
    
      /* Send alarm to node manager */
      get_timestamp_str (str);
      message = (char *) calloc ((strlen (alarmstr) + 
		         strlen (str) + 2), sizeof(char));
      if (message == NULL)
        logwriter ("CRITICAL: can't allocate memory for node manager alarm message, calloc()");
      else
        {
          snprintf (message, (strlen (alarmstr) + strlen (str) + 2), "%s %s", str, alarmstr);
          send_message_to_nodemgr (message);
          free (message);
        }
    }
#endif
}

void
logwriter (const char *log_entry)
{
  char str[STRBUF];

  get_timestamp_str (str);

  fprintf (logmode == LOG_STDOUT ? stdout : logfp,
	   "[%s][%d] %s\n", str, (int) getpid (), log_entry);
  fflush (logfp);
}
